// Version 6.00  07.04.2007 Markus Greither
// Copyright (c) 1991-2007 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efedit.h
    \brief Header for the function editor
*/

#ifndef __EFEDIT__
#define __EFEDIT__

#include "efbrcalc.h"
#include "efundo.h"

//! Constant for line breaks
#define LNBREAK "\r\n"

// ----------------------------------------------------------------------------
// TextSelection
// ----------------------------------------------------------------------------

//! Class for text marking
/*!
    Marked text is stored with line number and position of start and
    end of marked range.
*/
class TextSelection
{
 protected:
  //! Begin and end of selected range
  TextPos Anf,End;
  //! Clear selected range
  void Clear()
  {
    Anf.SetPos(0,1);
    End.SetPos(0,1);
    DblClk = 0;
  }
 public:
  //! Count for double click events, used for text selection
  int DblClk;
  //! Anf accessor
  const TextPos &GetAnf() const {return Anf;}
  //! End accessor
  const TextPos &GetEnd() const {return End;}
  //! Ctor
  TextSelection() {Clear();}
  //! Set new values
  void Set(const TextPos &A,const TextPos &E)
  {
    Anf = A;
    End = E;
  }
  //! Return true if multiple lines selected
  bool MehrereZeilen() {return (Anf.Zeile != End.Zeile);}
  //! Return true if any characters selected
  bool ZeichenMarkiert() {return (Anf.Pos != End.Pos);}
};

// ----------------------------------------------------------------------------
// SortedTextSelection
// ----------------------------------------------------------------------------

//! Extension to TextSelection with ability to sort
/*!
    Can sort text marking so that Zeile[1] > Zeile[0] and / or
    Pos[1] > Pos[0]. Necessary for comparisons.
*/
class SortedTextSelection : public TextSelection
{
 protected:
  //! Parent window
  wxWindow *m_Wnd;
  //! Flag for no macro output
  bool m_MacroOut;
  //! Clear
  void Clear()
  {
    TextSelection::Clear();
    Count = 2; // Sortierte Markierung
    AkPos[0].SetPos(0,1);
    AkPos[1].SetPos(0,1);
    AltPos.SetPos(0,1); // Anfang des aktuellen Bereiches
  }
 public:
  //! Ctor
  SortedTextSelection() : m_Wnd(0),m_MacroOut(true) {}
  //! Ctor
  SortedTextSelection(wxWindow *wnd) : m_Wnd(wnd),m_MacroOut(true) {}
  //! Ctor
  SortedTextSelection(const SortedTextSelection &other) : TextSelection(),m_Wnd(0),m_MacroOut(true)
  {
    *this = other;
  }
  //! Assignment operator, don't copy m_Wnd
  void operator = (const SortedTextSelection &other)
  {
    Set(other.GetAnf(),other.GetEnd(), false);
    AltPos = other.AltPos;
  }
  //! Set macro output flag, return previous value
  bool SetMacroOut(bool b)
  {
    bool tmp = m_MacroOut;
    m_MacroOut = b;
    return tmp;
  }
  //! Set new values
  void Set(const TextPos &A,const TextPos &E,bool recordmacro = true);
  //! Current position 0: before 1: within 2: after selection
  int Count;
  //! Sorted position (AkPos[0] < AkPos[1])
  TextPos AkPos[2];
  //! Cached old end position for repaint
  TextPos AltPos;
  //! Sort AkPos
  void SortPos()
  {
    Count = 0;
    if ((Anf.Zeile < End.Zeile) ||
        !MehrereZeilen() && (Anf.Pos <= End.Pos))
    {
      AkPos[0] = Anf;
      AkPos[1] = End;
    }
    else
    {
      AkPos[0] = End;
      AkPos[1] = Anf;
    }
  }
};

// ----------------------------------------------------------------------------
// FormatTextSelection
// ----------------------------------------------------------------------------

//! Extents text marking with format
class FormatTextSelection : public TextSelection
{
 public:
  //! Current text format
  int Format;
};

// ----------------------------------------------------------------------------
// ToolWindow
// ----------------------------------------------------------------------------

//! Class for Tooltip-Style windows
class ToolWindow : public wxWindow
{
  //! Text field
  wxStaticText *sttxt;

 public:
  //! Ctor
  ToolWindow(wxWindow *parent);
  //! Set Label and resize the window
  void SetText(const MGPStr &label);
  //! Paint this window
  void OnPaint(wxPaintEvent &event);
  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// EditTimer
// ----------------------------------------------------------------------------

class TEFEdit;

//! Class for update events.
class EditTimer : public wxTimer
{
  //! Editor for Timer events
  TEFEdit *Editor;
 public:
  //! Ctor
  EditTimer(TEFEdit *Ed) : wxTimer(),Editor(Ed) {}
  //! Overloaded notification function
  void Notify();
};

// ----------------------------------------------------------------------------
// TEFEdit
// ----------------------------------------------------------------------------

//! Text editor for formatted functions and comments
/*
    This editor emulates most of the WIN32 EDIT / wxTextCtrl functionality
    plus does automatic function formatting. Some text highlighting and
    error position reporting is also included.
    For cursor line position changes, the parent window is sent an event
    in order to do calculation updates if necessary.
    Modification information is available for each line.
    Multi-page printing support is available.
*/
class TEFEdit : public TEFTextCalc
{
  friend class EditTimer;
 public:
  //! Definition of edit clipboard format name
  static const wxChar *EditClipFormat;
  //! Style options
  enum EditMode
  {
    DialogStyle =   1,
    EquationStyle = 2,
    FunktionStyle = 4
  };
  //! View options
  enum ViewModes
  {
    ViewBruchDarst =    1,
    ViewHideDiags  =    2,
    ViewMarkLabels =    4,
    ViewKlammerDarst =  8,
    ViewScrollTrack =  16
  };

 protected:
  enum EditConsts
  {
    //! Line number increment per line
    COUNTSPERLINE = 2,
    //! X start position for text output in editor
    X0 =           10,
    //! Number of colors for bracket level coloring
    AnzFarbe =      7
  };
  //! Color array for bracket level coloring
  static const long FarbArr[AnzFarbe];
  //! true if left click and text selected
  bool DragMode;
  //! true if characters are inserted, not overwritten
  bool InsertMode;
  //! Flags for Display
  int ViewOptions;
  //! true if current printer setting is color
  bool ColorPrint;
  //! Style
  EditMode m_Mode;
  //! If true paint() redraws screen, otherwise redraws AltZeile to End.Zeile
  bool UpdateAll;
  //! Modify and mouse capture flags
  bool modified,Capt;
  //! true if SetFont() is called the first time
  bool FirstFont;
  //! Current selection for SetSelectionFormat()
  FormatTextSelection SelMark;
  //! Idle timer
  EditTimer IdleTimer;
  //! Undo / Redo data
  UndoData Rueck;
  //! MidZ position for bracked size
  int LastMidZ;
  //! Current selection
  SortedTextSelection m_Markiert;
  //! Scroll info (First visible line, max. visible lines)
  int FirstVisZeile,MaxZeilen;
  //! X begin position
  int XAnf;
  //! Pixels from top to first visible line
  int FirstVisPos;
  //! Length in pixels of current line
  int AkLen;
  //! Number of line with caret at the moment when idle timer was started
  int timerlin,aktimerlin;
  //! Editor text lines
  BruchArrObj BrZ;
  //! Function parser
  ParserObj Parser;
  //! String buffer for LockBuffer() and UnlockBuffer()
  MGPStr Buffer;
  //! Mouse position of current event
  wxPoint Pt;
  //! Mouse position of left button down event
  wxPoint DownPos;
  //! Right pointing arrow cursor for line selection
  wxCursor *RechtsCursor,*IBeamCursor;
  //! Visible window rectangle in pixels
  int VisXMax,VisYMax;
  //! Printer page orientation, true for landscape, false for portrait
  bool SeitenAusrichtung;
  //! Printer page size
  int PrntBreite,PrntLaenge;
  //! Printer page border sizes
  wxPoint PrntTL,PrntBR;
  //! Value window
  ToolWindow *ValueWin;
  //! Mouse position at the time of ValueWin creation
  wxPoint ValueWinPos;
  //! Helpers for popup menu
  bool MenuCursor,DelWin;
  //! Stopwatch for timing of scrolling
  wxStopWatch m_sw;

 public:
  //! Ctor
  TEFEdit() : TEFTextCalc(),IdleTimer(0),RechtsCursor(0)
  {
  }
  //! Ctor
  TEFEdit(wxWindow *parent,wxWindowID id,const wxPoint &pos,const wxSize &size,
          long style,const wxString &title,
          const NumListFormat &LDV,
          EditMode DialMode = TEFEdit::FunktionStyle);
  //! Dtor
  virtual ~TEFEdit();
  //! General event handling for macros
  bool ProcessEvent(wxEvent& event);
  //! Paint routine for print equations window
  void Paint(wxDC& dc,bool Prntng,wxSize& size);
  //! Paint routine for screen and printer
  void Paint(wxDC &dc,bool erase,wxRect &rect,bool forcenoborder = false);
  //! Add child window
  void AddChild(wxWindowBase *child);
  //! Delete child window
  void RemoveChild(wxWindowBase *child);
  //! Delete text selection
  bool DeleteSelection();
  //! Return text selection
  void GetSelection(unsigned int &A,unsigned int &E);
  //! Set text selection
  bool SetSelection(unsigned int A,unsigned int E,bool Scroll = true);
  //! Return number of first visible line
  int GetFirstVisibleLine() {return FirstVisZeile-1;}
  //! Return start position of line
  unsigned int GetLineIndex(int line);
  //! Return line number of position pos
  int GetLineFromPos(int pos);
  //! Insert text
  void Insert(const _TCHAR *L,bool Scroll = true,
              bool getundo = true,bool recordmacro = true);
  //! Insert page break
  void InsertPageBreak();
  //! Send modified event to view
  void SendModified();
  //! Return modified state
  bool IsModified() {return modified;}
  //! Clear modified state
  void ClearModify()
  {
    modified = false;
    Rueck.Clear();
  }
  //! Return text selection
  TextSelection &GetSelection() {return m_Markiert;}
  //! Return line string
  const MGPStr &GetLine(int N);
  //! Get line count
  unsigned int GetNumLines() {return (unsigned int)BrZ.GetMaxEintrag();}
  //! Return the length of line number Z
  int GetLineLength(int Z)
  {
    BrZ.SetAkEint(Z+1);
    return BrZ.AkZ().FZeile.Length();
  }
  //! Get total text length
  unsigned int GetTextLen();
  //! Cut current selection to clipboard
  void Cut();
  //! Copy current selection to clipboard
  void Copy(bool addbrackets = false);
  //! Paste clipboard data
  bool Paste();
  //! Clear editor
  void ClearEditor();
  //! Scroll in x and y direction
  int Scroll(int x,int y);
  //! Return current scroll position
  wxSize GetScrollSize()
  {
    wxSize size;
    size.x = XAnf;
    size.y = AbstVonAnf(FirstVisZeile,FHeight);
    return size;
  }
  //! Free Buffer
  void UnlockBuffer(const _TCHAR *)
  {
    Buffer.SetStr(wxT(""));
  }
  //! Copy all text into Buffer
  const _TCHAR *LockBuffer(int = 0);
  //! Search for given text
  int Search(int startPos, const _TCHAR *text,
             bool caseSensitive,bool wholeWord, bool up);
  //! Calculate first line of next page
  int NextPg(int fh,int &i,bool *Visible = 0);
  //! Print page
  void PaintPage(wxDC &DC,int Page);
  //! Set up document for printing
  int SetPrintDocSize(wxDC &dc);
  //! Set up document for screen
  void SetScreenDocSize();
  //! Format current selection
  void SetSelectionFormat(unsigned char Format,unsigned char H = 0);
  //! Set view option
  void SetViewOption(bool enable,ViewModes option,bool setmodified = true);
  //! Get view option
  bool GetViewOption(ViewModes option) {return (ViewOptions & option) != 0;}
  //! Return the editor mode
  TEFEdit::EditMode GetEditorMode() {return m_Mode;}
  //! Undo or redo last operation
  void UndoRedo(bool redo);
  //! Set page orientation for print
  void SetSeitenAusrichtung(bool Quer)
  {
    modified |= (SeitenAusrichtung != Quer);
    SeitenAusrichtung = Quer;
  }
  //! Return page orientation state
  bool GetSeitenAusrichtung()
  {
    return SeitenAusrichtung != 0;
  }
  //! Set font
  bool SetFont(const wxFont &font);
  //! Set zoom factor 1.0 == 100%
  void SetZoom(float Z);
  //! Return modified state for given line
  bool IsLineModified(int lin)
  {
    return BrZ.IsModified(lin);
  }
  //! Clear modified state for given line
  void ClearLineModified(int lin)
  {
    BrZ.ClearModified(lin);
  }
  //! Clear modified state for all lines
  void ClearLineModified();
  //! Set modified state for lines in SetFus
  void SetLineModified(bool set,vector<bool> &SetFus)
  {
    BrZ.SetModified(set,SetFus);
  }
  //! Set print settings
  void SetPrintSetup(int br,int h,wxPoint tplft,wxPoint btmrgt,
                     bool Or,bool col);
  //! Return print settings
  void GetPrintSetup(int &br,int &h,wxPoint &tplft,wxPoint &btmrgt,
                     bool &Or,bool &col);
  //! Check whether there are child windows on current print page
  bool VisibleChildren(int Zeile,int DrckFVisZeile,
                       int DrckLVisZeile,int FH);
  //! Respond to child window resize/move
  void OnChildSizeMove(wxCommandEvent &);
  //! Respond to past event
  void OnEditPaste(wxCommandEvent &);
  //! Respond to copy event
  void OnEditCopy(wxCommandEvent &);
  //! Respond to copy event, add brackets around text
  void OnEditCopyBracketText(wxCommandEvent &);
  //! Respond to spreadsheet copy event
  void OnSpreadCopy(wxCommandEvent &);
  //! Respond to cut event
  void OnEditCut(wxCommandEvent &);
  //! Respond to undo event
  void OnEditUndo(wxCommandEvent &);
  //! Respond to redo event
  void OnEditRedo(wxCommandEvent &);
  //! Enable / disable copy event
  void CeEditCopy(wxUpdateUIEvent& ce)
  {
    ce.Enable((m_Markiert.GetAnf().Pos != m_Markiert.GetEnd().Pos) ||
              (m_Markiert.GetAnf().Zeile != m_Markiert.GetEnd().Zeile));
  }
  //! Enable / disable paste event
  void CeEditPaste(wxUpdateUIEvent& ce);
  //! Enable / disable undo event
  void CeEditUndo(wxUpdateUIEvent& ce);
  //! Enable / disable redo event
  void CeEditRedo(wxUpdateUIEvent& ce);
  //! Respond to font bold event
  void OnFontBold(wxCommandEvent &)
  {
    SetSelectionFormat(TEFTextCalc::Bold);
  }
  //! Respond to font italic event
  void OnFontItalic(wxCommandEvent &)
  {
    SetSelectionFormat(TEFTextCalc::Italic);
  }
  //! Respond to font underline event
  void OnFontUnderline(wxCommandEvent &)
  {
    SetSelectionFormat(TEFTextCalc::Underline);
  }
  //! Respond to font strikeout event
  void OnFontStrikeout(wxCommandEvent &)
  {
    SetSelectionFormat(TEFTextCalc::Strikeout);
  }
  //! Respond to font height event
  void OnFontHeight(wxCommandEvent &);
  //! Enable / disable font format events
  void CeFontFormat(wxUpdateUIEvent &en)
  {
    BrZ.SetAkEint(m_Markiert.GetEnd().Zeile);
    en.Enable((FindFocus() == this) && (BrZ.AkZ().Commentline) &&
              (BrZ.AkZ().FZeile.Length() > 0));
  }
  //! Check bruch view menu entries
  void CeBruchDarst(wxUpdateUIEvent &ce)
  {
    ce.Enable(FindFocus() == this);
    ce.Check(ViewOptions & ViewBruchDarst);
  }
  //! Enable / disable bracket show menu entries
  void CeKlammerDarst(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && (ViewOptions & ViewBruchDarst));
    ce.Check((ViewOptions & ViewKlammerDarst) &&
             (ViewOptions & ViewBruchDarst));
  }
  //! Check diagram visible menu entries
  void CeHideDiags(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && (m_Mode == FunktionStyle));
    ce.Check((ViewOptions & ViewHideDiags) != 0);
  }
  //! Enable / disable mark labels menu entries
  void CeMarkLabels(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && (ViewOptions & ViewBruchDarst));
    ce.Check((ViewOptions & ViewMarkLabels) != 0);
  }
  //! Enable / disable spreadsheet copy
  void CeSpreadCopy(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this));
  }
  //! Respond to hide diags event
  void OnHideDiags(wxCommandEvent &event)
  {
    SetViewOption(event.IsChecked(),ViewHideDiags);
  }
  //! Responds to bruch view event
  void OnBruchDarst(wxCommandEvent &event)
  {
    SetViewOption(event.IsChecked(),ViewBruchDarst);
  }
  //! Responds to bracket show event
  void OnKlammerDarst(wxCommandEvent &event)
  {
    SetViewOption(event.IsChecked(),ViewKlammerDarst);
  }
  //! Responds to mark labels event
  void OnMarkLabels(wxCommandEvent &event)
  {
    SetViewOption(event.IsChecked(),ViewMarkLabels);
  }
  //! Responds to set modified event
  void OnSetModified(wxCommandEvent &)
  {
    modified = true;
  }
  //! Responds to drag text event
  unsigned int OnDragText(wxCoord x,wxCoord y,bool keepselection);
  //! Shows tooltip-Style window with string value
  void ShowValueWindow(const MGPStr &value);
  //! Focus handling
  void SetFocus();
 protected:
  //! Extends the current selection to next higher bracket combination
  void ExtendSelection(bool DoubleClick);
  //! Gets the next bracket combination left and right from current positions
  void GetNewSelection(unsigned int &Anf,unsigned int &End);
  //! Checks whether to mark labels
  bool GetMarkLabels()
  {
    return !BrZ.AkZ().FormatError &&
           (ViewOptions & ViewMarkLabels) && (ViewOptions & ViewBruchDarst);
  }
  //! Checks whether to display this text in red
  bool GetRedText(unsigned int Pos,unsigned int f,unsigned int ZPos)
  {
    return BrZ.AkZ().FormatError && !BrZ.AkZ().Commentline &&
           (Pos >= f-(ZPos-BrZ.AkZ().FehlerPos)) && !(m_Mode == DialogStyle);
  }
  //! Check whether character x is a marker
  bool IstBMarker(char x)
  {
    return (x == _TCHAR(TEFTextCalc::PosMarker)) || 
           (x == _TCHAR(TEFTextCalc::FontMarker));
  }
  //! Check whether current line is a page break
  bool IstPagebreak()
  {
    return (_TCHAR(BrZ.AkZ().FZeile[0]) == _TCHAR(TEFTextCalc::FontMarker)) &&
           (_TCHAR(BrZ.AkZ().FZeile[1]) == _TCHAR(TEFTextCalc::PageMarker));
  }
  //! Calculate distance from document top in pixels for line i
  int AbstVonAnf(int i,int Fh);
  //! Draw child windows for printing
  void PaintChildren(wxDC &dc,int AAnfX,
                     int DrckFVisZeile,int DrckLVisZeile,wxRect &cliprect);
  //! Check whether given child window is within printable range
  bool IsChildVisible(wxWindow *child,int DrckFVisZeile,
                      int DrckLVisZeile,int AAnfX,int FH,wxRect &rect);
  //! Show / hide child windows
  void ShowDiags(bool hide);
  //! Draw formatted text, using Color Bracket or BigBracked functions
  bool DrawFormattedText(wxPoint p, const _TCHAR *string, int count,
                         wxSize &size,int &LZaehler,int &KlZaehler);
  //! Draw text with coloured bracket levels
  bool DrawColorBracketText(wxPoint p, const _TCHAR *string, int count,
                            wxSize &size,int &LZaehler,int &KlZaehler);
  //! Draw text with brackets in current line height
  bool DrawBigBracketText(wxPoint p, const _TCHAR *string, int count,
                          wxSize &size,int &LZaehler);
  //! Draw text, mark labels and comments
  bool DrawAndMarkText(wxPoint p, const _TCHAR *string, int count,
                       wxSize &size,int &LZaehler);
  //! Draw text (low level), one piece for screen, char by char for print / clpb
  void DrawText(wxPoint p, const _TCHAR *string, int count, wxSize &size);
  //! Redraw the current line
  void RePaintAkZeile();
  //! Paint current line
  int PaintZeile(int &AkPos);
  //! Determine type of current line (function / comment), recalculate Bruch
  bool EvalZeile(int &Len);
  //! Draws parts of equations or comment lines
  void TextSectionOut(const _TCHAR *Zeile,unsigned int &Pos,
                        unsigned int &XPos,unsigned int &YPos,
                        unsigned int &ZPos,int &LZaehler);
  //! Draws parts of equations or comment lines (no LZaehler)
  void TextSectionOut(const _TCHAR *Zeile,unsigned int &Pos,
                        unsigned int &XPos,unsigned int &YPos,
                        unsigned int &ZPos)
  {
    int LZaehler = 0;
    TextSectionOut(Zeile,Pos,XPos,YPos,ZPos,LZaehler);
  }
  //! Set current text color, depending of error state
  void SetTextColor(bool highlight,bool redtext = false);
  //! Draw separation lines for print pages
  void DrawPageLine(wxDC &dc);
  //! Get document size in VisXMax, VisYMax
  void GetDocSize(int AFHeight = 0,float Zoom = 1);
  //! Zoom child windows
  void ZoomChildren(int AXAnf,int AltFVisPos,float OldZoom);
  //! Update scroll positions after zoom change
  void UpdateScrollPos(int AXAnf);
  //! Set new first visible line (vertical scroll pos), returns delta y
  int SetFirstVisZeile(int FV,bool ScrollAlleZeilen = false);
  //! Returns the outline rectangle of the current line
  bool GetAkZeileRect(wxRect &arect);
  //! Draw carent position in case we don't have the focus
  void DrawCaretPosMarker(int x,int y,bool erase = false);
  //! Refresh the screen, recalculate document size
  void EFInvalidate(bool Erase = true);
  //! Set scrollbar thumb size and position
  void SetScrollbar(int orient,int pos,int PageSize,
                    int range,bool refresh = true);
  //! Copy current selection (Enanced Metafile) into clipboard
  void MetafileToClipboard(wxDataObjectComposite &cmp);
  //! Copy current selection (text format) into clipboard
  bool SelToClipboard(bool erase,wxDataObjectComposite &cmp,bool addbrackets = false);
  //! Retrieve currently selected text, delete it if erase = true
  bool GetSelText(MGPStr &Zeile,bool erase,bool getundo = true);
  //! Calculate the height in pixels of current line
  inline int NextAkPos(int FH = 0)
  {
    return (BrZ.AkZ().ObZ[1]+BrZ.AkZ().UntZ[1]+COUNTSPERLINE)*(FH ?
                                                               FH : FHeight);
  }
  //! Returns the line height in pixels of this comment line
  int GetCommentLineHeight(const MGPStr &Zeile);
  //! Recalculate lines from A to E
  void UpdateZeilen(int A,int E,bool modified = false);
  //! Calculate number of lines for page up
  int PgUpZeilen(int AkZeile);
  //! Calculate number of lines for page down
  int PgDownZeilen(int AkZeile);
  //! Returns the last currently fully visible line
  int LastVisZeile(int AkZeile);
  //! Scroll the window so that the caret is in view
  bool ScrollToCaret();
  //! Scroll window by xdel and ydel
  void ScrollWindowPos(int xdel,int ydel);
  //! Determine the line / character position of the mouse cursor
  bool MousePos(bool clicked,int &Zeile,
                unsigned int &Pos,bool keepselection = false);
  //! Determine the line of the mouse cursor
  bool MouseLine(int &PixelPos);
  //! Set up caret
  void InitCaret();
  //! Set caret position to x, y
  void SetCaretPos(int x,int y);
  //! Set caret position if current position is right and / or down from caret
  void EvalCaretPos(unsigned int &XPos,unsigned int &YPos,unsigned int &ZPos);
  //! Get document size, update scrollbars
  void SetDocSize(bool getdocsize = true,int DieseZeile = 0,bool updatescrollbars = true);
  //! Scroll the window if caret dragged out of the window with left button down
  bool ScrollWithCursor(wxPoint Pt);
  //! Responds to timer event
  void OnTimer();
  //! Responds to key strokes
  void OnKeyDown(wxKeyEvent &event);
  //! Responds to character input
  void OnChar(wxKeyEvent &event);
  //! Responds to right button up event
  void OnRButtonUp(wxMouseEvent &event);
  //! Responds to left button down event
  void OnLButtonDown(wxMouseEvent &event);
  //! Responds to left button doubleclick event
  void OnLButtonDblClk(wxMouseEvent &event);
  //! Responds to left button up event
  void OnLButtonUp(wxMouseEvent &event);
  //! Responds to mouse leave event
  void OnMouseLeave(wxMouseEvent &event);
  //! Responds to mouse move event
  void OnMouseMove(wxMouseEvent &event);
  //! Responds to set cursor event
  void OnSetCursor(wxSetCursorEvent &event);
  //! Responds to paint event
  void OnPaint(wxPaintEvent &event);
  //! Responds to set focus event
  void OnSetFocus(wxFocusEvent &event);
  //! Responds to kill focus event
  void OnKillFocus(wxFocusEvent &event);
  //! Scroll window in vertical direction
  void OnVScroll(WXTYPE ScrollCode,int line);
  //! Scrollwindow in horizontal direction
  void OnHScroll(WXTYPE ScrollCode,int line);
  //! Responds to scroll event
  void OnScroll(wxScrollWinEvent &event);
  //! Responds to size event
  void OnSize(wxSizeEvent &event);
  //! Responds to mouse wheel event
  void OnMousewheel(wxMouseEvent &event);
  //! Responds to delete window event
  void OnDelEdit(wxCommandEvent &hp);
  //! Responds to extend selection event
  void OnExtendSelection(wxCommandEvent &event);
  //! Respond to select all functions event
  void OnMarkAll(wxCommandEvent &);
  //! Respond to macro set font height event
  void OnMacroFontHeight(wxCommandEvent &event);
  //! Respond to macro set selection event
  void OnMacroSetSelection(wxCommandEvent &event);
  //! Respond to macro text insert
  void OnMacroInsert(wxCommandEvent &event);
  //! Respond to macro delete key
  void OnMacroDelete(wxCommandEvent &event);
  //! Respond to macro backspace key
  void OnMacroBackspace(wxCommandEvent &event);
  //! Respond to macro return key
  void OnMacroReturn(wxCommandEvent &event);
  //! Respond to macro child window resize/move
  void OnMacroResizeChild(wxCommandEvent &event);

  DECLARE_EVENT_TABLE()
  DECLARE_DYNAMIC_CLASS(TEFEdit)
};

#endif

