// Version 6.06  29.11.2009 Markus Greither
// Copyright (c) 1991-2009 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efview.h
    \brief Header for the view class of the Doc/View framework
*/

#ifndef __EFVIEW__
#define __EFVIEW__

#include "wx/docview.h"

// ----------------------------------------------------------------------------
// TEFView
// ----------------------------------------------------------------------------

//! Easy Funktion view class
/*
    Manages the children of the main function window (TEFEdit).
*/
class TEFView: public wxView
{
 public:
  //! Ctor
  TEFView();
  //! Dtor
  virtual ~TEFView();
  //! General event handling, prevents recursive wxCommandEvent loops
  bool ProcessEvent(wxEvent& event);
  //! Set up main function editor
  TEFEditSearch *CreateEditor(wxFrame *parent);
  //! Create a printout object
  wxPrintout* CreatePrintout();
  //! Respond to print preview events
  void OnPreview(wxCommandEvent &event);
  //! Respond to print events
  void OnPrintFocusWindow(wxCommandEvent &event);
  //! Respond to page setup events
  void OnPageSetup(wxCommandEvent& event);
  //! Respond to view update events
  void OnUpdate(wxView *sender,wxObject * /*hint*/);
  //! Respond to view create events
  bool OnCreate(wxDocument *doc, long flags);
  //! Respond to view close events
  bool OnClose(bool deleteWindow = TRUE);
  //! Respond to view draw events
  void OnDraw(wxDC *);
  //! Respond to view activate events
  void OnActivateView(bool activate, wxView *activeView, wxView *deactiveView);
  //! Respond to diagram modified events
  void OnSetDiagModified(wxCommandEvent &) {LastCalcDiagW = 0;};
  //! Set currently active diagram window
  void SetAkDiag(TEFGraphControl *GG);
  //! Copy all functions in active editor into Fu
  void GetAkEditFunktionen(FunctionArrObj *Fu,bool all = true)
  {
    GetFunktionen(Fu,all,AkEdit ? AkEdit : Editor);
  }
  //! Solve equations from currently active Editor for interactive solving
  bool SolveEquations();
  //! Interactively multiply/divide operand
  void ShowInteractiveSolveResult(FunctionArrObj &Fu,SolverObj *Solver);
  //! Insert string in currently active function editor
  void InsertString(const _TCHAR *zeile,int Anf,int End);
  //! Get the current functions name without []
  void AkFuNameOhneKlammern(MGPStr &nm);
  //! Activate current diagram window
  void SetActiveDiag(bool Act)
  {
    if (AkDiagW)
      AkDiagW->SetActive(Act);
  };
  //! Return main function editor
  inline TEFEditSearch *GetEditor()
  {
    return Editor;
  };
  //! Set main function editor functions
  void SetEditorFunk();
  //! Read all functions from doc into Fu, ignore entries of already loaded submodules list
  bool GetSubmodulesFromDoc(TEFDocument *doc,FunctionArrObj *Fu,MGPStrSet &submodules,int &Count);
  //! Load functions from edit and defined submodules
  bool GetFunctionsIncludingModules(FunctionArrObj *Fu,bool AlleFkts,TEFEdit *edit = 0);
  //! Load functions from edit
  bool GetFunktionen(FunctionArrObj *Funktion,bool AlleFkts,TEFEdit *edit = 0);
  //! Parse line of text from editor and determine name, parameters and term
  void ParseFunktion(const MGPStr &Str,CpxObj &Funk, TEFEdit *edit);
  //! Display error message
  void Fehlermeldung(ParserObj::FehlerRec &Fehler,bool FehlerPos,TEFEdit *edit = 0);
  //! Get data from various windows for file save
  void VnCommit(DiagDataPtArrObj &DiagDataPts);
  //! Returns modified state
  inline bool IsModified() const
  {
    if (Editor)
      return Editor->IsModified();
    else
      return false;
  }
  //! Clear modified state
  void ClearModify();
  //! Load data into various windows
  bool LoadData(bool NewDiag);

 protected:
  //! print data, to remember settings during the session
  wxPrintData *g_printData;
  //! page setup data
  wxPageSetupData* g_pageSetupData;
  //! Main function editor font
  FontType EditFont;
  //! Graph calculation progress dialog
  wxDialog *BerDlg;
  //! Label entry dialog
  wxDialog *LabelEntryDlg;
  //! Function array
  FunctionArrObj *Funktion;
  //! Document
  TEFDocument *EFDoc;
  //! Main function editor
  TEFEditSearch *Editor;
  //! Currently active diagram window
  TEFGraphControl *AkDiagW,*LastCalcDiagW;
  //! Currently active edit window
  TEFEdit *AkEdit;

  //! Renumber all child windows of Editor
  void RenumberChildWindows();
  //! Counts the number of diagrams in the main editor
  int GetDiagCount();
  //! Paste graphics as child of main function editor
  void PasteBild();
  //! Setup page size and orientation for given editor
  void SetEditorPrintSetup(TEFEdit *edit,bool init = false,wxDC *dc = 0);
  //! Create new diagram window, setup modes and colour
  void CreateAkDiag();
  //! Create new diagram window
  void NeuAkDiag(DiagDataArrObj *DDA,bool usescrollposy = false);
  //! Get editor window font
  void GetEditorFont(TEFEdit *ed,FontType &EditFont);
  //! Set editor window font
  void SetEditorFont(TEFEdit *ed,const FontType &EditFont);
  //! Create new equations window
  void NeuEquations(const wxPoint &pos,const wxSize &size,
                    const _TCHAR *data = 0,const FontType *Font = 0,
                    const BildschOptObj *BOpt = 0,
                    const wxSize *Selection = 0,
                    const wxSize *ScrollPos = 0);
  //! Put [] around all functions names
  void AddKlammern(FunctionArrObj *Fu,int Anf = 0,int End = 0);
  //! Calculate graph
  bool Berechnen(bool Update,bool AkFunk,TEFGraphControl *ThisDiag = 0);
  //! Update all graphs
  void UpdateAll(bool UpdateOnDiagChange);
  //! Respond to editor set focus event
  void OnEditSetFocus(wxCommandEvent &event);
  //! Respond to set active diagram window event
  void OnSetAkDiag(wxCommandEvent &event);
  //! Respond to set modified event
  void OnSetModified(wxCommandEvent &event);
  //! Respond to get functions event
  void OnGetFunktions(wxCommandEvent &event);
  //! Respond to calculate value event
  void OnWertBerechnen(wxCommandEvent &event);
  //! Respond to key press event
  void OnKeyDown(unsigned int key, unsigned int repeatCount, unsigned int flags);
  //! Respond to new equations window event
  void OnNeuEquations(wxCommandEvent &);
  //! Respond to new diagram window event
  void OnNeuAkDiag(wxCommandEvent &);
  //! Respond to zoom changed event
  void OnZoom(wxCommandEvent &);
  //! Respond to insert page break event
  void OnPageBreak(wxCommandEvent &);
  //! Respond to set editor font event
  void OnEditFont(wxCommandEvent &);
  //! Respond to label entry dialog event
  void OnLabelEntry(wxCommandEvent &);
  //! Respond to single value dialog event
  void OnEinzWert(wxCommandEvent &);
  //! Toggle single value panel button
  void CeEinzWert(wxUpdateUIEvent &event);
  //! Respond to submodules dialog event
  void OnSubmodules(wxCommandEvent &);
  //! Toggle submodules panel button
  void CeSubmodules(wxUpdateUIEvent &event);
  //! Respond to update graph event
  void OnUpdateGraph(wxCommandEvent &);
  //! Respond to update all graphs event
  void OnUpdateAll(wxCommandEvent &);
  //! Respond to new graph event
  void OnBerechnen(wxCommandEvent &);
  //! Enable/disable events that apply only to function editor
  void CeEditor(wxUpdateUIEvent &event)
  {
    event.Enable(wxWindow::FindFocus() == Editor);
  }
  //! Respond to solve functions event
  void OnUmformen(wxCommandEvent &);
  //! Respond to approximation function event
  void OnNaeherung(wxCommandEvent &);
  //! Respond to function selection panel event
  void OnFunktionen(wxCommandEvent &);
  //! Toggle function selection panel button
  void CeFunktionen(wxUpdateUIEvent &event);
  //! Respond to interactive solver panel event
  void OnInteractiveSolver(wxCommandEvent &);
  //! Toggle interactive solver panel button
  void CeInteractiveSolver(wxUpdateUIEvent &event);
  //! Respond to paste event
  void OnEditPaste(wxCommandEvent &event);
  //! Enable / Disable paste event
  void CeEditPaste(wxUpdateUIEvent &event);
  //! Respond to event that needs to be passed to focus window
  void OnFocusWindowEvent(wxCommandEvent &event)
  {
    wxWindow *wnd = wxWindow::FindFocus();
    TEFEdit *edit = wxDynamicCast(wnd,TEFEdit);
    if (wnd && Editor &&
        ((Editor->GetChildren().IndexOf(wnd) != wxNOT_FOUND) ||
         (wnd == Editor) || edit))
    {
      wnd->GetEventHandler()->ProcessEvent(event);
    }
  }
  //! Enable / disable event specific to focus window
  void CeFocusWindowEvent(wxUpdateUIEvent &event)
  {
    wxWindow *wnd = wxWindow::FindFocus();
    TEFEdit *edit = wxDynamicCast(wnd,TEFEdit);
    if (wnd && Editor &&
        ((Editor->GetChildren().IndexOf(wnd) != wxNOT_FOUND) ||
         (wnd == Editor) || edit))
      wnd->GetEventHandler()->ProcessEvent(event);
    else
      event.Enable(false);
  }
  //! Enable / disable file save event
  void CeSave(wxUpdateUIEvent &event);
  //! Enable / disable events that apply to all editors
  void CeAnyEdit(wxUpdateUIEvent &event)
  {
    event.Enable(wxDynamicCast(wxWindow::FindFocus(),TEFEdit) != 0);
  }
  //! Enable / disable delete current diagram window event
  void CeDelDiag(wxUpdateUIEvent &ce);
  //! Send this document per eMail
  void OnSendTo(wxCommandEvent &);
  //! Open program homepage
  void OnHomepage(wxCommandEvent &);
  //! Open program internet forum
  void OnForum(wxCommandEvent &);
  //! Send feedback email
  void OnFeedback(wxCommandEvent &);
  //! Replay macro
  void OnPlayMacro(wxCommandEvent &);
  //! Replay macro
  void OnRecordMacro(wxCommandEvent &);
  //! Replay macro
  void OnEndMacro(wxCommandEvent &);
  //! Enable macro stop
  void CeMacroActive(wxUpdateUIEvent &event);
  //! Respond to delete current window event
  void OnDelWindow(wxCommandEvent &event);
  //! Enable macro recording/playback
  void CeMacroInactive(wxUpdateUIEvent &event);
  //! Execute next macro event
  void OnMacro(wxCommandEvent &);
  //! Clear the modified flats
  void OnClearModified(wxCommandEvent &);
  //! Idle event handling
  void OnRenumberChildren(wxCommandEvent &);
  //! Macro set font handler
  void OnMacroEditFont(wxCommandEvent &event);

  DECLARE_EVENT_TABLE()
  DECLARE_DYNAMIC_CLASS(TEFView)
};

#endif

