// Version 6.05 20.07.2008 Markus Greither
// Copyright (c) 1991-2008 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efedit.cpp
    \brief Source code for the function editor internal code
*/


#include "efprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#ifdef __DMC__
#include <typeinfo.h>
#endif
#include <wx/wx.h>
#include <wx/resizec.h>

#include <wx/clipbrd.h>

#include <complex>
#include <limits>
#include <climits>
#include <set>
#include <functional>
#if defined(__MINGW32__)
#include <wx/msw/winundef.h>
#endif

using namespace std;

#include "efdefs.h"
#include "mg.h"
#include "cpxpsr.h"
#include "diagramm.h"
#include "efgemein.h"
#include "ef.h"
#include "efresid.h"
#include "eftabcon.h"
#include "efedit.h"

#endif

#include <wx/caret.h>

#ifdef __WXGTK__
#include "xpm/toright.xbm"
#endif

#include "efeddnd.h" // local include because problem for prec. headers

// ----------------------------------------------------------------------------
// ToolWindow
// ----------------------------------------------------------------------------

ToolWindow::ToolWindow(wxWindow *parent)
 : wxWindow(parent,-1,wxPoint(10,10),wxSize(10,10))
{
  sttxt = new wxStaticText(this,-1,wxT(""),wxPoint(2,2),GetClientSize());
#ifdef __DMC__
  SetBackgroundColour(wxColour(0xFF,0xFF,0xE1));
  sttxt->SetBackgroundColour(wxColour(0xFF,0xFF,0xE1));
#else
  SetBackgroundColour(wxSystemSettings::GetColour(wxSYS_COLOUR_INFOBK));
  sttxt->SetBackgroundColour(wxSystemSettings::GetColour(wxSYS_COLOUR_INFOBK));
#endif
}

void ToolWindow::SetText(const MGPStr &label)
{
  sttxt->SetLabel(label.c_str());
  int x,y;
  GetSize(&x,&y); // prevent window shrink
  if (x < sttxt->GetSize().x+4)
    x = sttxt->GetSize().x+4;
  if (y < sttxt->GetSize().y+4)
    y = sttxt->GetSize().y+4;
  SetSize(x,y);
  Raise();
  Show(true);
}

void ToolWindow::OnPaint(wxPaintEvent &)
{
  wxPaintDC dc(this);
  wxBrush OrgBr = dc.GetBrush();
  wxBrush br(GetBackgroundColour(),wxSOLID);
  dc.SetBrush(br);
  dc.DrawRectangle(0,0,GetClientSize().x,GetClientSize().y);
  dc.SetBrush(OrgBr);
}

BEGIN_EVENT_TABLE(ToolWindow,wxWindow)
  EVT_PAINT(ToolWindow::OnPaint)
END_EVENT_TABLE()

// ----------------------------------------------------------------------------
// TEFEdit
// ----------------------------------------------------------------------------

const wxChar *TEFEdit::EditClipFormat = wxT("EasyFunktion Text Data 1.0");

const long TEFEdit::FarbArr[] =
{
  0x00000000L,0x000000FFL,0x00009F00L,0x00FF0000L,
  0x00BF00BFL,0x00AFAF00L,0x0000AFAFL
};

IMPLEMENT_DYNAMIC_CLASS(TEFEdit,TEFTextCalc)

TEFEdit::TEFEdit(wxWindow *parent,wxWindowID id,const wxPoint &pos,
                 const wxSize &size,long style,const wxString &title,
                 const NumListFormat &LDV,EditMode Md)
 : TEFTextCalc(parent,id,pos,size,style,title),
   InsertMode(true),m_Mode(Md),FirstFont(true),
   IdleTimer(this),m_Markiert(this),Parser(LDV),MenuCursor(false)
{
  SetDropTarget(new EFTextDropTarget(this));
#ifdef __WXMSW__
  RechtsCursor = new wxCursor(wxT("IDC_PFEILNACHRECHTS"));
#else
  RechtsCursor = new wxCursor(toright,32,32,32,7);
#endif
  IBeamCursor = new wxCursor(wxCURSOR_IBEAM);
  DelWin = false;
  ViewOptions = ViewBruchDarst+ViewMarkLabels;
  SeitenAusrichtung = 0; // portrait page
  PrntBreite = 0;
  PrntLaenge = 0;
  ClearEditor();
  wxFont font(10,wxROMAN,wxNORMAL,wxNORMAL,false,wxT("Times New Roman"));
  SetFont(font);
  SetScrollbar(wxVERTICAL,0,0,FHeight*COUNTSPERLINE,true);
  SetScrollbar(wxHORIZONTAL,0,0,GetClientSize().GetWidth(),true);
  SelMark.Format = TEFTextCalc::Normal;
  PrntTL = wxPoint(15,15);
  PrntBR = wxPoint(15,15);
  ValueWin = 0;
}

TEFEdit::~TEFEdit()
{
  delete RechtsCursor;
  delete IBeamCursor;
  if (ValueWin)
  {
    ValueWin->Destroy();
    ValueWin = 0;
  }
}

void TEFEdit::SetZoom(float Z)
{
  int AXAnf = XAnf;
  float Zoom1 = Zoom;
  OrgZoom = Z;
  AveChrW = int(FontMetric.OrgAveCharW*OrgZoom);
  FHeight = int((FontMetric.OrgHeight*FSizeFak*OrgZoom+1)/COUNTSPERLINE);
  if (FHeight < 1)
    FHeight = 1; // Minimum font height
  Zoom = float(FHeight)/
         float(int((FontMetric.OrgHeight*FSizeFak+1)/COUNTSPERLINE));
  float RelZoom = Zoom/Zoom1;
  XAnf = int(XAnf*RelZoom);
  MaxZeilen = (GetClientSize().GetHeight())/(FHeight*COUNTSPERLINE);
  GetDC();
  UpdateZeilen(1,BrZ.GetMaxEintrag());
  DelDC();
  if (Zoom != Zoom1)
  {
    int AltFVisPos = FirstVisPos;
    FirstVisPos = AbstVonAnf(FirstVisZeile,FHeight); // Neuer Unsichtbarer Bereich
    ZoomChildren(AXAnf,AltFVisPos,Zoom1);
  }
  if (m_Mode == FunktionStyle)
    ScrollToCaret();
  EFInvalidate();
}

void TEFEdit::ZoomChildren(int AXAnf,int AltFVisPos,float OldZoom)
{

  wxWindowListNode *Child = GetChildren().GetFirst();
  while (Child)
  {
    wxResizeableControl *tst = wxDynamicCast(Child->GetData(),wxResizeableControl);
    if (tst)
    {
#ifdef __WXGTK__
      // Under GTK, window position changes are not always caught (Scrolling)
      wxRect rect = ((wxWindow *)Child->GetData())->GetRect();
      tst->GetZoomData().Move(rect.x,rect.y);
#endif
      tst->GetZoomData().SetZoomRect((wxWindow *)Child->GetData(),Zoom,
                                     int((AXAnf-X0)/OldZoom),int(AltFVisPos/OldZoom),
                                     int((XAnf-X0)/Zoom),int(FirstVisPos/Zoom));
    }
    Child = Child->GetNext();
  }
}

/*! Determines height of invisible space on top,
    corrects child window positions
*/
void TEFEdit::UpdateScrollPos(int AXAnf)
{
  int AltFVisPos = FirstVisPos;
  FirstVisPos = AbstVonAnf(FirstVisZeile,FHeight); // New invisible space
  if ((FirstVisPos != AltFVisPos) || (AXAnf != XAnf))
  { // if invisible space or scroll pos changed
    wxWindowListNode *Child = GetChildren().GetFirst();
    while (Child)
    { // Correct window positions
      wxRect rect = ((wxWindow *)Child->GetData())->GetRect();
      // Determine new window position
      rect.y += (+AltFVisPos);
      rect.x += (+AXAnf-X0);
      rect.x += (-XAnf+X0);
      rect.y += (-FirstVisPos);
      ((wxWindow *)Child->GetData())->SetSize(rect);
      Child = Child->GetNext();
    }
  }
}

// Set new first visible line
int TEFEdit::SetFirstVisZeile(int FV,bool ScrollAlleZeilen)
{
  int AkPos;
  if (VisYMax < BrZ.GetMaxEintrag())
    GetDocSize();
  if (FV < 1)
    FirstVisZeile = 1;
  else
  if ((FV > VisYMax) && !ScrollAlleZeilen)
    FirstVisZeile = VisYMax;
  else
    FirstVisZeile = FV;
  AkPos = AbstVonAnf(FirstVisZeile,FHeight);
  if (AkPos != FirstVisPos)
  {
    ScrollWindowPos(0,FirstVisPos-AkPos);
    FirstVisPos = AkPos;
  }
  return AkPos-FirstVisPos;
}

void TEFEdit::ScrollWindowPos(int delx,int dely)
{
  HideCaret();
  CaretYPos += dely;
  CaretXPos += delx;
  wxWindow::ScrollWindow(delx,dely);
  GetCaret()->Move(CaretXPos,CaretYPos);
  ShowCaret();
}

// Sets/deletes view options
void TEFEdit::SetViewOption(bool enable,ViewModes option,bool setmodified)
{
  if (enable != ((ViewOptions & option) > 0))
  {
    if (enable)
      ViewOptions |= option;
    else
      ViewOptions &= ~option;
    if (option == ViewHideDiags)
      ShowDiags(enable);
    else
    {
      GetDC();
      UpdateZeilen(1,BrZ.GetMaxEintrag());
      DelDC();
      UpdateScrollPos(XAnf);
      ScrollToCaret();
      EFInvalidate();
    }
    if (setmodified)
    {
      modified = true;
      SendModified();
    }
  }
}

// Determine most right position and lowest line in document
void TEFEdit::GetDocSize(int AFHeight,float Zm)
{
  wxSize GesSize(0,0);
  int AbstB = AbstVonAnf(FirstVisZeile,AFHeight ? AFHeight : FHeight);
  if (!(ViewOptions & ViewHideDiags))
  {
    wxWindowListNode *child = GetChildren().GetFirst();
    while (child)
    {
      wxWindow *chw = wxDynamicCast(child->GetData(),wxResizeableControl);
      if (chw)
      {
        wxRect rect = chw->GetRect();
        rect.x += XAnf-X0;
        rect.y += AbstB;
        if (rect.GetBottom() > GesSize.y)
          GesSize.y = rect.GetBottom();
        if (rect.GetRight()+1 > GesSize.x)
          GesSize.x = rect.GetRight()+1;
      }
      child = child->GetNext();
    }
  }
  int AkPos = 0;
  BrZ.Begin();
  while (BrZ.NextEint())
    AkPos += NextAkPos(); // compute text length
  int AkLine = BrZ.GetMaxEintrag();
  // if lowest child window below last line, add virtual blank lines
  while (AkPos < float(GesSize.y)/Zm)
  {
    AkLine++;
    AkPos += FHeight*COUNTSPERLINE;
  }
  VisYMax = AkLine;
  if (int(AkLen+AveChrW) > float(GesSize.x)/Zm)
    VisXMax = AkLen+AveChrW;
  else
    VisXMax = int(float(GesSize.x)/Zm);
}

// Determine height of space on top of line i
int TEFEdit::AbstVonAnf(int i,int Fh)
{
  int Abst = 0;
  int OrgPos = BrZ.AkEintNummer();
  BrZ.Begin();
  while (BrZ.NextEint() && (BrZ.AkEintNummer() < i))
  {
    Abst += NextAkPos(Fh);
  }
  if (i > BrZ.GetMaxEintrag())
    Abst += (i-BrZ.GetMaxEintrag())*Fh*COUNTSPERLINE;
  BrZ.SetAkEint(OrgPos);
  return Abst;
}

void TEFEdit::ShowDiags(bool hide)
{
  wxWindowListNode *child = GetChildren().GetFirst();
  while (child)
  {
    ((wxWindow *)child->GetData())->Show(!hide);
    child = child->GetNext();
  }
}

// Verify current line if comment or calculates view
bool TEFEdit::EvalZeile(int &Len)
{
  int Anf;

  Len = 0;
  BrZ.AkZ().BZeile.SetStr(wxT(""));
  Anf = -1;
  BrZ.AkZ().FehlerPos = 0;
  BrZ.AkZ().FormatError = false;
  BrZ.AkZ().Commentline = (BrZ.AkZ().FZeile[0] == ' ') ||
                          (BrZ.AkZ().FZeile.Length() !=
                           BrZ.AkZ().FZeile.FLength()) ||
                          (BrZ.AkZ().FZeile.FLength() == 0);
  if (!BrZ.AkZ().Commentline)
  {
    // if Dialog edit, evaluate term on both sides of =
    if ((m_Mode == EquationStyle) ||
        (m_Mode == DialogStyle) ||
        (BrZ.AkZ().FZeile[0] == '['))
      Anf = 0;
    else
    { // output function name first
      bool Fehler = false;
      Anf = Parser.IstGleichPos(BrZ.AkZ().FZeile.c_str(),&Fehler);
      if (Fehler)
        BrZ.AkZ().FormatError = true;
      if (Anf != -1)
      { // '=' found
        Anf++;
        while (BrZ.AkZ().FZeile[Anf] == ' ')
          Anf++; // determine end of last blank char
        Len = EFTabbedTextExtent(BrZ.AkZ().FZeile.c_str(),Anf,
                                 GetMarkLabels()).x;
        BrZ.AkZ().BZeile.AddNChar(BrZ.AkZ().FZeile.c_str(),Anf);
      }
      else // '=' not found
        Anf = 0;
    }
    BrZ.AkZ().ObZ.SetValue(1,0);
    BrZ.AkZ().UntZ.SetValue(1,0);
    BrZ.AkZ().MidZ.SetValue(1,0);
    BrZ.AkZ().X.SetValue(0,Len);
  }
  if (!BrZ.AkZ().FormatError && !BrZ.AkZ().Commentline)
  { // if no error, evaluate view
    BruchCalcObj BrCalc(BrZ.AkZ(),Parser,this,
                        ViewOptions & ViewKlammerDarst,
                        ViewOptions & ViewMarkLabels);
    Len += BrCalc.FindBruch(&BrZ.AkZ().FZeile(Anf),Len);
  }
  if (!(ViewOptions & ViewBruchDarst) ||
      BrZ.AkZ().FormatError || BrZ.AkZ().Commentline)
    Len = EFTabbedTextExtent(BrZ.AkZ().FZeile.c_str(),BrZ.AkZ().FZeile.Length(),
                             GetMarkLabels()).x;
  BrZ.AkZ().LineLength = Len;
  if (!BrZ.AkZ().FormatError && !BrZ.AkZ().Commentline)
    return true;
  else
  { // error occured
    BrZ.AkZ().BZeile.SetStr(wxT(""));
    int H = 0;
    if (BrZ.AkZ().Commentline)
      H = GetCommentLineHeight(BrZ.AkZ().FZeile);
    BrZ.AkZ().ObZ.SetValue(1,H);
    BrZ.AkZ().UntZ.SetValue(1,0);
    return false;
  }
}

int TEFEdit::GetCommentLineHeight(const MGPStr &Zeile)
{
  unsigned int i = 0;
  int Height = 0;
  while (i < Zeile.Length())
  {
    if (Zeile[i] == _TCHAR(TEFTextCalc::FontMarker))
    {
      i++;
      if ((Zeile[i] != _TCHAR(TEFTextCalc::PageMarker)) &&
          (Zeile[i] & TEFTextCalc::FontHeight))
      {
        i++;
        int tmp = (/*unsigned*/_TCHAR)Zeile[i];
        int H = tmp/FontMetric.FPoints*COUNTSPERLINE;
        if (tmp % FontMetric.FPoints > FontMetric.FPoints/2)
          H += COUNTSPERLINE;
        else
        if (tmp % FontMetric.FPoints)
          H += COUNTSPERLINE/2;
        if (H)
        {
          if (H > COUNTSPERLINE)
            H -= COUNTSPERLINE;
          else
            H = 0;
        }
        if (H > Height)
          Height = H;
      }
    }
    i++;
  }
  return Height;
}

// search current Caret position
void TEFEdit::EvalCaretPos(unsigned int &XPos,unsigned int &YPos,
                           unsigned int &ZPos)
{
  if (Capt)
  { // left mouse button pressed
    if (m_Markiert.GetEnd().Zeile == 0)
    { // mouse position not found yet
      if ((m_Markiert.GetEnd().Zeile == 0) && ((Pt.x <= int(XPos-XAnf+X0)) &&
          (Pt.y <= int(YPos+FHeight*COUNTSPERLINE)) ||
          (BrZ.AkZ().FZeile(ZPos) == 0)))
      { // first mouse position below and right of Pt or end of line
     
        TextPos NeuAnf = m_Markiert.GetAnf(),NeuEnd = m_Markiert.GetEnd();
        NeuEnd.Zeile = BrZ.AkEintNummer();
        NeuEnd.Pos = ZPos;
        m_Markiert.Set(NeuAnf,NeuEnd);
        SetCaretPos(XPos-XAnf+X0-1,YPos);
      }
    }
  }
  else
  if ((m_Markiert.GetEnd().Zeile == BrZ.AkEintNummer()) &&
      (m_Markiert.GetEnd().Pos == ZPos) && (m_Markiert.Count < 2))
  {
    if ((m_Markiert.Count == 0) || (m_Markiert.AkPos[0].Pos < m_Markiert.GetEnd().Pos) ||
        (m_Markiert.AkPos[0].Zeile < BrZ.AkEintNummer()))
    { // only if TextOut didn't increase first position
      SetCaretPos(XPos-XAnf+X0-1,YPos);
      m_Markiert.Count++;
      if (!m_Markiert.ZeichenMarkiert() && !m_Markiert.MehrereZeilen())
        m_Markiert.Count++;
    }
  }
  else
  if ((m_Markiert.Count == 0) &&

      (m_Markiert.AkPos[0].Zeile <= BrZ.AkEintNummer()) &&
      (BrZ.AkZ().FZeile.Length() == 0))
    m_Markiert.Count++; // for selected blank lines
  else
  if ((m_Markiert.Count < 2) &&
      (m_Markiert.AkPos[m_Markiert.Count].Zeile == BrZ.AkEintNummer()) &&
      (m_Markiert.AkPos[m_Markiert.Count].Pos == ZPos))
    m_Markiert.Count++;
}

// Format the marked section of comment line
void TEFEdit::SetSelectionFormat(unsigned char Format,unsigned char H)
{
  MGPStr Zeile;

  if (m_Markiert.ZeichenMarkiert() || m_Markiert.MehrereZeilen())
  { // marked section available
    if ((SelMark.GetEnd() == m_Markiert.GetEnd()) &&
        (SelMark.GetAnf() == m_Markiert.GetAnf()))
    { // same section than with last call
      if (Format & TEFTextCalc::FontHeight)
        Format |= SelMark.Format;
      else
      if ((Format & SelMark.Format) == Format)
        Format = (unsigned char)((SelMark.Format & ~Format) | 
                                 TEFTextCalc::Normal);
      else
        Format |= (unsigned char)(SelMark.Format);
    }
    SelMark.Format = Format; // store current selection
    SelMark.Set(m_Markiert.GetAnf(),m_Markiert.GetEnd());
    m_Markiert.SortPos();
    BrZ.SetAkEint(m_Markiert.AkPos[0].Zeile);
    unsigned short AFormat = 0;
    unsigned short Format1;
    if (BrZ.AkZ().FZeile(0) == ' ')
    { // format only comment lines
      if (m_Markiert.AkPos[0].Pos < BrZ.AkZ().FZeile.FLength())
      { // if selection starts before the end of first selected line
        FormatStrObj tmp;
        tmp.SetStr(BrZ.AkZ().FZeile);
        tmp.ShortAfterMarker(m_Markiert.AkPos[0].Pos);
        Format1 = tmp.GetTextStr(Zeile); // get format before first position
        if ((Format1 > 0) && (m_Markiert.AkPos[0].Pos != 0))
          AFormat = Format1;
        else
          AFormat = TEFTextCalc::Normal;
        tmp.SetStr(BrZ.AkZ().FZeile);
        if (Format != AFormat)
        { // if not regular font, add format
          if (H) // if new font height specified, use it
            BrZ.AkZ().FZeile.InsFontMarker(m_Markiert.AkPos[0].Pos,Format,H);
          else
          if (Format1 & 0xFF00) // otherwise use old font height
            BrZ.AkZ().FZeile.InsFontMarker(m_Markiert.AkPos[0].Pos,Format |
                                           TEFTextCalc::FontHeight,Format1 >> 8);
          else
            BrZ.AkZ().FZeile.InsFontMarker(m_Markiert.AkPos[0].Pos,Format,0);
        }
        else
          BrZ.AkZ().FZeile.DelFontMarker(m_Markiert.AkPos[0].Pos);
        if (m_Markiert.AkPos[0].Zeile == m_Markiert.AkPos[1].Zeile)
          tmp.ShortAfterMarker(m_Markiert.AkPos[1].Pos);
        Format1 = tmp.GetTextStr(Zeile); // get format before second position
        if (Format1 > 0)
          AFormat = Format1;
        else
          AFormat = TEFTextCalc::Normal;
        BrZ.AkZ().FZeile.GetTextStr(Zeile); // get text string
        if (m_Markiert.AkPos[1].Zeile == m_Markiert.AkPos[0].Zeile)
        { // if same line only remove format for middle part
          Zeile.ShortStr(m_Markiert.AkPos[1].Pos);
          // append formatted rest string
          Zeile.AddStr(BrZ.AkZ().FZeile.FromMarkPos(m_Markiert.AkPos[1].Pos));
        }
        // append
        BrZ.AkZ().FZeile.FAddAtStr(m_Markiert.AkPos[0].Pos,
                                   &Zeile[m_Markiert.AkPos[0].Pos]);
      }
    }
    else
    { // no comment line
      BrZ.AkZ().FZeile.GetTextStr(Zeile);
      BrZ.AkZ().FZeile.SetStr(Zeile); // remove formats
    }
    if (m_Markiert.AkPos[1].Zeile == m_Markiert.AkPos[0].Zeile)
    { // selection in 1 line
      if ((m_Markiert.AkPos[1].Pos < BrZ.AkZ().FZeile.FLength()) &&
          (AFormat != Format) && (BrZ.AkZ().FZeile(0) == ' '))
      { // if second position not after last char
        if (*BrZ.AkZ().FZeile.FromMarkPos(m_Markiert.AkPos[1].Pos) !=
            _TCHAR(TEFTextCalc::FontMarker))
        { // if font marker not yet there
          BrZ.AkZ().FZeile.InsFontMarker(m_Markiert.AkPos[1].Pos,
                                         AFormat & 0xFF,AFormat >> 8);
        }
      }
    }
    else
    { // selection over more than one line
      while (BrZ.AkEintNummer() < m_Markiert.AkPos[1].Zeile-1)
      { // format intermediate lines
        BrZ.NextEint();
        Format1 = BrZ.AkZ().FZeile.GetTextStr(Zeile);
        BrZ.AkZ().FZeile.SetStr(Zeile);
        if ((Format != TEFTextCalc::Normal) && (BrZ.AkZ().FZeile(0) == ' '))
        { // if not regular font, insert format
          if (H) // if font height specified, use it
            BrZ.AkZ().FZeile.InsFontMarker(m_Markiert.AkPos[0].Pos,Format,H);
          else
          if (Format1 & 0xFF00) // otherwise use eventual old font height
            BrZ.AkZ().FZeile.InsFontMarker(0,Format |
                                           TEFTextCalc::FontHeight,
                                           Format1 >> 8);
          else
            BrZ.AkZ().FZeile.InsFontMarker(0,Format,0);
        }
      }
      if (BrZ.AkEintNummer() < m_Markiert.AkPos[1].Zeile) BrZ.NextEint();
      if (BrZ.AkZ().FZeile(0) == ' ')
      { // only format comment lines
        if (m_Markiert.AkPos[1].Pos > 0)
        { // format last line if end position > 0
          FormatStrObj tmp;
          tmp.SetStr(BrZ.AkZ().FZeile);
          tmp.ShortAfterMarker(m_Markiert.AkPos[1].Pos);
          Format1 = tmp.GetTextStr(Zeile); // get format before second pos
          if (Format1 > 0) AFormat = Format1;
          else AFormat = TEFTextCalc::Normal;
          BrZ.AkZ().FZeile.GetTextStr(Zeile); // get text line
          // append formatted rest string
          Zeile.AddAtStr(m_Markiert.AkPos[1].Pos,
                         &BrZ.AkZ().FZeile(m_Markiert.AkPos[1].Pos));
          BrZ.AkZ().FZeile.SetStr(Zeile);
          if ((m_Markiert.AkPos[1].Pos < Zeile.Length()) && (AFormat != Format))
          { // if not after last char, insert format
            if (*BrZ.AkZ().FZeile.FromMarkPos(m_Markiert.AkPos[1].Pos) !=
                _TCHAR(TEFTextCalc::FontMarker))
            { // if not font marker is there already
              BrZ.AkZ().FZeile.InsFontMarker(m_Markiert.AkPos[1].Pos,AFormat &
                                                     0xFF,AFormat >> 8);
            }
          }
          if (Format != TEFTextCalc::Normal)
          { // if not regular font insert format
            if (H) // use font height if available
              BrZ.AkZ().FZeile.InsFontMarker(m_Markiert.AkPos[0].Pos,Format,H);
            else
            // otherwise use eventual old font height
            if (Format1 & 0xFF00)
              BrZ.AkZ().FZeile.InsFontMarker(0,Format |
                                             TEFTextCalc::FontHeight,
                                             Format1 >> 8);
            else
              BrZ.AkZ().FZeile.InsFontMarker(0,Format,0);
          }
        }
      }
      else
      { // no comment line
        BrZ.AkZ().FZeile.GetTextStr(Zeile); // remove formats
        BrZ.AkZ().FZeile.SetStr(Zeile);
      }
    }
  }
  int Min = -BrZ.AkZ().ObZ[1],Max = BrZ.AkZ().UntZ[1];
  GetDC();
  UpdateZeilen(m_Markiert.AkPos[0].Zeile,m_Markiert.AkPos[1].Zeile,true);
  DelDC();
  BrZ.SetAkEint(m_Markiert.AkPos[0].Zeile);
  if ((m_Markiert.AkPos[0].Zeile == m_Markiert.AkPos[1].Zeile) &&
      (-BrZ.AkZ().ObZ[1] == Min) && (BrZ.AkZ().UntZ[1] == Max))
    RePaintAkZeile();
  else EFInvalidate();
  modified = true;
  SendModified();
}

// Update line formatting for lines from A to E
void TEFEdit::UpdateZeilen(int A,int E,bool modified)
{
  if ((A == 1) && (E == BrZ.GetMaxEintrag()))
    AkLen = 0; // Update maximum line length
  int Len;
  for (int i = A;i <= E;i++)
  {
    BrZ.SetAkEint(i);
    if (modified)
      BrZ.AkZ().Modified = true;
    EvalZeile(Len);
    if (Len > AkLen)
      AkLen = Len;
  }
}

// Determine document size and updates the scrollbars
void TEFEdit::SetDocSize(bool getdocsize,int DieseZeile,bool updatescrollbars)
{
  if (!DieseZeile)
    DieseZeile = FirstVisZeile; // Default
  if (getdocsize)
    GetDocSize();

  if (updatescrollbars)
  {
    int vsize = LastVisZeile(DieseZeile)-DieseZeile+1;
    SetScrollbar(wxVERTICAL,DieseZeile-1,vsize,VisYMax,true);
  
    int hsize = GetClientSize().x-X0;
    SetScrollbar(wxHORIZONTAL,XAnf,hsize,VisXMax,true);
  }  
}

void TEFEdit::SetScrollbar(int orient, int Pos,
                           int Page,int range, bool refresh)
{
/*  if (orient == wxVERTICAL) // Debug-Code for Scrollbar
  {
    wxScreenDC dc;
    dc.SetTextBackground(WINDOWCOLOUR);
    dc.SetBackgroundMode(wxSOLID);
    char Z[200];
    sprintf(Z,"%i %i %i     ",Pos,Page,range);
    dc.DrawText(Z,1,1);
  }*/
  static bool SetScrollActive = false;
  if (SetScrollActive)
  {
    Refresh();
    return;
  }
  SetScrollActive = true;
  if (Page+Pos > range)
    Page = range-Pos;
  wxWindow::SetScrollbar(orient,Pos,Page,range,refresh);
  SetScrollActive = false;
}

// Determines rectangle of current line
bool TEFEdit::GetAkZeileRect(wxRect &arect)
{
  arect = wxRect(0,0,GetClientSize().x,GetClientSize().y);
  int AkPos = 0;
  if (FirstVisZeile <= m_Markiert.GetEnd().Zeile)
  {
    BrZ.SetAkEint(FirstVisZeile-1);
    while ((BrZ.AkEintNummer() < m_Markiert.GetEnd().Zeile-1) && BrZ.NextEint())
      AkPos += NextAkPos();
    if (AkPos <= GetClientSize().y)
      arect.y = AkPos; // begin of rectangle with current line
    else 
      arect.y = GetClientSize().y; // limit to client size
    BrZ.NextEint();
    arect.height = NextAkPos(); // end of rectangle with current line
    if (arect.y+arect.height > GetClientSize().y)
      arect.height = GetClientSize().y-arect.y; // auf ClSize begrenzen
    return arect.height;
  }
  else
    return false;
}

// Scroll the client area to caret
bool TEFEdit::ScrollToCaret()
{
  bool scrolled = false;
  if (m_Markiert.GetEnd().Zeile < FirstVisZeile)
  {
    SetFirstVisZeile(m_Markiert.GetEnd().Zeile);
    scrolled = true;
  }
  else
  if (LastVisZeile(FirstVisZeile) < m_Markiert.GetEnd().Zeile)
  {
    int y = 0;
    BrZ.SetAkEint(LastVisZeile(FirstVisZeile)+1);
    do
    { // calculate lines from LastVisZeile+1 to and including End.Zeile
      y += NextAkPos();
    }
    while (BrZ.NextEint() && (BrZ.AkEintNummer() <= m_Markiert.GetEnd().Zeile));
    int i = 0;
    BrZ.SetAkEint(FirstVisZeile);
    do
    { // count lines from above until enough room for lines below
      i += NextAkPos();
    }
    while (BrZ.NextEint() && (i < y));
    SetFirstVisZeile(BrZ.AkEintNummer());
    scrolled = true;
  }
  int x = 0;
  GetDC();
  BrZ.SetAkEint(m_Markiert.GetEnd().Zeile);
  m_Markiert.SortPos();
  // make sure that output is above of client area
  int i = (-BrZ.AkZ().ObZ[1]-BrZ.AkZ().UntZ[1]-COUNTSPERLINE)*FHeight;
  if ((m_Markiert.AkPos[0].Zeile < m_Markiert.GetEnd().Zeile) &&
      (m_Markiert.AkPos[1].Zeile >= m_Markiert.GetEnd().Zeile))
    m_Markiert.Count = 1;
  else m_Markiert.Count = 0;
  HideCaret();
  // calculate x position of caret without moving the caret
  int len = PaintZeile(i);
  ShowCaret();
  DelDC();
  if (len < GetClientSize().GetWidth()-X0-AveChrW)
    x = XAnf; // if whole string fits in window, scroll to 0
  else
  if (CaretXPos+1 < X0)
    x = X0-CaretXPos-1; // if caret position left of window
  else
  // if caret position right of window
  if (CaretXPos+1+AveChrW > GetClientSize().GetWidth())
    x = GetClientSize().GetWidth()-CaretXPos-1-AveChrW;
  else
  // if scrolled farther than necessary
  if (len-GetClientSize().GetWidth()+X0+AveChrW < XAnf)
    x = XAnf-len+GetClientSize().GetWidth()-X0-AveChrW;
  if (x != 0)
  {
    XAnf -= x;
    ScrollWindowPos(x,0);
    scrolled = true;
  }
  return scrolled;
}

void TEFEdit::EFInvalidate(bool Erase)
{
  SetDocSize();
  Refresh(Erase);
}

// Calculates the number of lines that are scrolled for PAGEUP
int TEFEdit::PgUpZeilen(int AkZeile)
{
  int i = 0,f = AkZeile;
  while ((f > 1) && (i < COUNTSPERLINE*MaxZeilen))
  {
    f--;
    if (f <= BrZ.GetMaxEintrag())
    {
      BrZ.SetAkEint(f);
      i += NextAkPos(1);
    }
    else
      i += COUNTSPERLINE;
  }
  if (i > COUNTSPERLINE*MaxZeilen)
    f++;
  return AkZeile-f;
}

// Calculates the last visible line
int TEFEdit::LastVisZeile(int AkZeile)
{
  int i = 0,f = AkZeile-1;
  do
  { // determine the last completely visible line
    f++;
    if (f <= BrZ.GetMaxEintrag())
    {
      BrZ.SetAkEint(f);
      i += NextAkPos(1);
    }
    else
      i += COUNTSPERLINE;
  }
  while (i < COUNTSPERLINE*MaxZeilen);
  if (i > COUNTSPERLINE*MaxZeilen)
    return f-1;
  else
    return f;
}

// Calculate number of lines that are scrolled for PAGEDOWN
int TEFEdit::PgDownZeilen(int AkZeile)
{
  return LastVisZeile(FirstVisZeile)-AkZeile+1;
}

// Determine selected range, erases it if erase = true
bool TEFEdit::GetSelText(MGPStr &Zeile,bool erase,bool getundo)
{
  Zeile.SetStr(wxT(""));
  m_Markiert.SortPos();
  if (m_Markiert.ZeichenMarkiert() || m_Markiert.MehrereZeilen())
  { // selection available
    BrZ.SetAkEint(m_Markiert.AkPos[0].Zeile);
    if (m_Markiert.AkPos[0].Zeile != m_Markiert.AkPos[1].Zeile)
    { // multi line selection
      int n = m_Markiert.AkPos[1].Zeile-m_Markiert.AkPos[0].Zeile;
      int f;
      if (m_Markiert.AkPos[0].Pos > 0)
      {
        Zeile.AddStr(BrZ.AkZ().FZeile.FromMarkPos(m_Markiert.AkPos[0].Pos));
        if (erase)
        {
          // if not first char short line
          BrZ.AkZ().FZeile.FShortStr(m_Markiert.AkPos[0].Pos);
          f = m_Markiert.AkPos[0].Zeile+1;
        }
        else
          f = m_Markiert.AkPos[0].Zeile;
      }
      else
      {
        Zeile.AddStr(BrZ.AkZ().FZeile);
        if (erase)
          BrZ.LoeschAkEint(); // otherwise delete line
        f = m_Markiert.AkPos[0].Zeile;
      }
      Zeile.AddStr(wxT(LNBREAK)); // append line break
      for (int i = 1;i < n;i++)
      {
        if (erase)
          BrZ.SetAkEint(f);
        else
          BrZ.SetAkEint(f+i);
        Zeile.AddStr(BrZ.AkZ().FZeile);
        Zeile.AddStr(wxT(LNBREAK));
        if (erase)
          BrZ.LoeschAkEint(); // delete following lines
      }
      if (erase)
        BrZ.SetAkEint(f);
      else
        BrZ.SetAkEint(f+n);
      if (m_Markiert.AkPos[1].Pos > 0)
      {
        Zeile.AddNChar(BrZ.AkZ().FZeile,
                       BrZ.AkZ().FZeile.ModCharPos(m_Markiert.AkPos[1].Pos));
        if (erase)
          BrZ.AkZ().FZeile.FDelNChar(0,m_Markiert.AkPos[1].Pos);
      }
      if (erase && (f > m_Markiert.AkPos[0].Zeile))
      { // first line truncated -> append
        const _TCHAR *tmp = BrZ.AkZ().FZeile.c_str();
        BrZ.SetAkEint(m_Markiert.AkPos[0].Zeile);
        BrZ.AkZ().FZeile.AddStr(tmp);
        BrZ.NextEint();
        BrZ.LoeschAkEint();
      }
    }
    else
    {
      Zeile.AddNChar(&BrZ.AkZ().FZeile(m_Markiert.AkPos[0].Pos),
                     BrZ.AkZ().FZeile.ModLen(BrZ.AkZ().FZeile.ModCharPos(m_Markiert.AkPos[0].Pos),
                     m_Markiert.AkPos[1].Pos-m_Markiert.AkPos[0].Pos));
      if (erase)
        BrZ.AkZ().FZeile.FDelNChar(m_Markiert.AkPos[0].Pos,
                                   m_Markiert.AkPos[1].Pos-m_Markiert.AkPos[0].Pos);
    }
    if (erase)
    {
      TextPos NeuAnf,NeuEnd;

      NeuAnf = m_Markiert.AkPos[0];
      NeuEnd = m_Markiert.AkPos[0];
      m_Markiert.Set(NeuAnf,NeuEnd);
      if (getundo)
        Rueck.NewUndo(m_Markiert.GetEnd(),m_Markiert.GetEnd(),m_Markiert.GetEnd(),
                      Zeile);
      GetDC();
      UpdateZeilen(m_Markiert.GetEnd().Zeile,m_Markiert.GetEnd().Zeile,true);
      DelDC();
      UpdateScrollPos(XAnf);
    }
    return true;
  }
  return false;
}

// Insert page break
void TEFEdit::InsertPageBreak()
{
  TextPos NeuAnf = m_Markiert.GetAnf(),NeuEnd = m_Markiert.GetEnd();

  BrZ.SetAkEint(NeuEnd.Zeile);
  BrZ.InsAkEint();
  BrZ.AkZ().FZeile.InsFontMarker(0,TEFTextCalc::PageMarker,0);
  int AEndZeile = NeuEnd.Zeile;
  NeuEnd.Zeile++;
  NeuEnd.Pos = 0;
  Rueck.NewUndo(TextPos(0,AEndZeile),TextPos(0,NeuEnd.Zeile),
                TextPos(0,AEndZeile),wxT(""));
  NeuAnf = NeuEnd;
  // Do not set macro set selection, because out of order with CMD_PAGEBREAK
  m_Markiert.Set(NeuAnf,NeuEnd,false);
  GetDC();
  UpdateZeilen(AEndZeile,NeuEnd.Zeile,true);
  DelDC();
  modified = true;
  SendModified();
  UpdateScrollPos(XAnf);
  if (IsShown())
    ScrollToCaret();
  EFInvalidate();
}

void TEFEdit::SetCaretPos(int x,int y)
{
  CaretXPos = x;
  CaretYPos = y;
  if (FindFocus() == this)
  {
    if (GetCaret())
      GetCaret()->Move(x,y);
  }
  else
    DrawCaretPosMarker(x,y);
}

// Determine the line and char position of mouse cursor
bool TEFEdit::MousePos(bool notappend,int &Zeile,
                       unsigned int &Pos,bool keepselection)
{
  int i1;

  GetDC();
  int tmpy = Pt.y;
  MouseLine(i1);
  // recalculate line with mouse position
  SortedTextSelection TM = m_Markiert;
  bool oldmo = m_Markiert.SetMacroOut(false);

  TextPos NeuAnf = m_Markiert.GetAnf(),NeuEnd = m_Markiert.GetEnd();
  NeuEnd.Zeile = 0;
  NeuAnf.Zeile = 0;
  m_Markiert.Set(NeuAnf,NeuEnd,false);

  m_Markiert.SortPos();
  if ((m_Markiert.AkPos[0].Zeile < BrZ.AkEintNummer()) &&
      (m_Markiert.AkPos[1].Zeile >= BrZ.AkEintNummer()))
    m_Markiert.Count = 1;
  else
    m_Markiert.Count = 0;
  int tmpx = Pt.x;
  bool Links = false;
  if (Pt.x < X0-1-XAnf)
  {  // position on the left beside the text
    Links = true;
    tmpx = Pt.x;
    int tmpAZ = TM.GetAnf().Zeile;
    if (notappend)
      tmpAZ = BrZ.AkEintNummer();
    if (tmpAZ <= BrZ.AkEintNummer() && !keepselection)
      Pt.x = 32000; // mouse on the left beside the text -> select whole line
  }
  HideCaret();
  PaintZeile(i1);
  ShowCaret();
  DelDC();
  Pt.x = tmpx;
  Pt.y = tmpy;
  Zeile = m_Markiert.GetEnd().Zeile;
  Pos = m_Markiert.GetEnd().Pos;
  m_Markiert = TM;
  m_Markiert.SetMacroOut(oldmo);
  return Links;
}

bool TEFEdit::MouseLine(int &i1)
{
  int i = 0;
  bool ret = true;

  BrZ.SetAkEint(FirstVisZeile);
  if (Pt.y > GetClientSize().y)
  {
    Pt.y = GetClientSize().y;
    ret = false;
  }
  else
  if (Pt.y < 0)
  {
    Pt.y = 0;
    ret = false;
  }
  do
  { // calculate all lines from above until mouse position found
    i1 = i;
    i += NextAkPos();
  }
  while ((i < Pt.y) && BrZ.NextEint());
  return ret;
}

