// Version 6.06  19.01.2010 Markus Greither
// Copyright (c) 1991-2010 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efview.cpp
    \brief Source code for the view class of the Doc/View framework
*/


#include "efprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#include <wx/xrc/xmlres.h>

#ifndef WX_PRECOMP
#ifdef __DMC__
#include <typeinfo.h>
#endif
#include <wx/wx.h>

#include <wx/docview.h>
#include <wx/clipbrd.h>
#include <wx/spinctrl.h>
#include <wx/resizec.h>
#include <wx/image.h>

#include <cstring>
#include <set>
#include <functional>
#if defined(__MINGW32__)
#include <wx/msw/winundef.h>
#endif
#include <complex>
#include <limits>
#include <cstdio>

using namespace std;

#include "efdefs.h"
#include "mg.h"
#include "cpxpsr.h"
#include "diagramm.h"
#include "gauss.h"
#include "efgemein.h"
#include "efsolver.h"
#include "efedit.h"
#include "efeditsr.h"
#include "efdocum.h"
#include "efresid.h"
#include "ef.h"
#include "eftabcon.h"
#include "efgraphc.h"
#include "efeqatns.h"
#include "effnkclc.h"
#include "efview.h"
#include "efdlgs.h"
#include "efnmdlgs.h"
#include "efprint.h"

#endif

#include <wx/dcps.h>
#include <wx/xrc/xh_toolb.h>
#include <wx/xrc/xh_combo.h>
#include "efmacro.h"

#include <wx/print.h>
#include <wx/printdlg.h>
#include <wx/mstream.h>
#if defined(__BORLANDC__) || defined(__MINGW32__) || defined(__VISUALC__)
#include <wx/net/email.h>
#endif

#define WINDOWCOLOUR wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW)

// ----------------------------------------------------------------------------
// TEFView
// ----------------------------------------------------------------------------

IMPLEMENT_DYNAMIC_CLASS(TEFView, wxView)

TEFView::TEFView(): wxView()
{
  Editor = 0;
  AkDiagW = 0;
  AkEdit = 0;
  EFDoc = 0;
  g_printData = 0;
  g_pageSetupData = 0;
}

TEFView::~TEFView()
{
  delete LabelEntryDlg;
  delete g_printData;
  delete g_pageSetupData;
}

bool TEFView::ProcessEvent(wxEvent& event)
{
  wxWindow *wnd = wxWindow::FindFocus();
  if (wnd && (wnd != Editor) && Editor &&
      (Editor->GetChildren().IndexOf(wnd) == wxNOT_FOUND))
    wnd = 0;
  int id = (wnd ? wnd->GetId() : 0);

  // Lock macro recorder to suppress duplicate events
  bool oldlock = MyMacroRecorder::Get()->SetLock(true);

  bool ret = wxView::ProcessEvent(event);

  MyMacroRecorder::Get()->SetLock(oldlock);
  if (ret)
    MyMacroRecorder::Get()->AddEvent(event,id);

  wxKeyEvent *tmp = wxDynamicCast(&event,wxKeyEvent);
  if (tmp && tmp->AltDown())
    return false;

  if (!ret && AkDiagW && (event.GetEventType() == wxEVT_UPDATE_UI))
    ret = AkDiagW->GetEventHandler()->ProcessEvent(event); // AkDiagw default so that UPDATE_UI works

  if (!ret && ((event.GetEventType() == wxEVT_COMMAND_MENU_SELECTED) ||
               (event.GetEventType() == wxEVT_UPDATE_UI)))
  {
    TEFEdit *edit = wxDynamicCast(wxWindow::FindFocus(),TEFEdit);

    if (!ret && EFDoc && Editor)
      if (edit && (edit != Editor)) // special case edit (no direct child of Editor)
        ret = edit->GetEventHandler()->ProcessEvent(event); // Events also travel up to Editor
      else
      if (Editor->GetChildren().IndexOf(wxWindow::FindFocus()) != wxNOT_FOUND)
        ret = wxWindow::FindFocus()->GetEventHandler()->ProcessEvent(event); // Picture window or diag
      else
        ret = Editor->GetEventHandler()->ProcessEvent(event); // Make sure Editor gets events
  }
  return ret;
}

void TEFView::OnPageSetup(wxCommandEvent& /*event*/)
{
  (*g_pageSetupData) = *g_printData;
  wxSize psize;
  wxPoint tplft,btmrgt;
  bool Or,Col;
  Editor->GetPrintSetup(psize.x,psize.y,tplft,btmrgt,Or,Col);
  g_printData->SetOrientation(Or ? wxLANDSCAPE : wxPORTRAIT);
  g_pageSetupData->SetMarginTopLeft(tplft);
  g_pageSetupData->SetMarginBottomRight(btmrgt);

  wxPageSetupDialog pageSetupDialog(wxGetApp().GetTopWindow(), g_pageSetupData);
  pageSetupDialog.ShowModal();

  (*g_printData) = pageSetupDialog.GetPageSetupData().GetPrintData();
  (*g_pageSetupData) = pageSetupDialog.GetPageSetupData();
  SetEditorPrintSetup(Editor,false);
}

void TEFView::SetEditorPrintSetup(TEFEdit *edit,bool init,wxDC *dc)
{
  wxSize psize;
  if (init)
    g_printData->SetOrientation(EFDoc->SeiteQuer ? wxLANDSCAPE : wxPORTRAIT);
  wxDC *DC;
  if (dc)
    DC = dc;
  else
#ifdef __WXMSW__
    DC = new wxPrinterDC(*g_printData);
#else
    DC = new wxPostScriptDC(*g_printData);
#endif
  if (DC && DC->IsOk())
  {
    psize = DC->GetSize();
    wxSize sz = DC->GetPPI();
    psize.x = int(psize.x*25.4/sz.x);
    psize.y = int(psize.y*25.4/sz.y);
  }
  if (!init && (edit == Editor))
    EFDoc->SeiteQuer = (g_printData->GetOrientation() == wxLANDSCAPE);
  if (init)
  {
    g_pageSetupData->SetMarginTopLeft(wxPoint(15,15));
    g_pageSetupData->SetMarginBottomRight(wxPoint(15,15));
  }
  edit->SetPrintSetup(psize.x,psize.y,
                      (edit != Editor) ? wxPoint(15,15) :
                      g_pageSetupData->GetMarginTopLeft(),
                      (edit != Editor) ? wxPoint(15,15) :
                      g_pageSetupData->GetMarginBottomRight(),
                      (edit != Editor) ? wxPORTRAIT :
                      g_printData->GetOrientation() == wxLANDSCAPE,
                      g_printData->GetColour());
  if (!dc)
    delete DC;
}

void TEFView::OnPreview(wxCommandEvent&)
{
  wxWindow *win = Editor->FindFocus();
  TEFEquationsControl *eqct = wxDynamicCast(win,TEFEquationsControl);
  if (eqct)
    win = eqct->GetEditor();
  TEFEdit *edit = wxDynamicCast(win,TEFEdit);
  if (!edit && (win != AkDiagW))
    win = Editor;
  if (win == AkDiagW)
    g_printData->SetOrientation(!AkDiagW->ZweiDiags() ?
                                wxLANDSCAPE : wxPORTRAIT);
  wxPrintout *printout = CreatePrintout();
  if (printout)
  {
    // Pass two printout objects: for preview, and possible printing.
    wxPrintDialogData printDialogData(* g_printData);
    wxPrintPreview *preview = new wxPrintPreview(printout,
                                                 CreatePrintout(),
                                                 &printDialogData);

    wxPreviewFrame *frame = new wxPreviewFrame(
            preview,
            (wxFrame *)wxTheApp->GetTopWindow(),
            _("Print preview"),
            wxGetApp().GetTopWindow()->GetPosition(),
            wxGetApp().GetTopWindow()->GetSize());
    frame->Initialize();
    frame->Show(true);
  }
  g_printData->SetOrientation(EFDoc->SeiteQuer ? wxLANDSCAPE : wxPORTRAIT);
}

void TEFView::OnPrintFocusWindow(wxCommandEvent&)
{
  wxWindow *win = Editor->FindFocus();
  TEFEquationsControl *eqct = wxDynamicCast(win,TEFEquationsControl);
  if (eqct)
    win = eqct->GetEditor();
  TEFEdit *edit = wxDynamicCast(win,TEFEdit);
  if (!edit && (win != AkDiagW))
    win = Editor;
  if (win == AkDiagW)
    g_printData->SetOrientation(!AkDiagW->ZweiDiags() ?
                                wxLANDSCAPE :
                                wxPORTRAIT);
  wxPrintout *printout = CreatePrintout();
  if (printout)
  {
    wxPrintDialogData pddata(*g_printData);
    wxPrinter printer(&pddata);
    wxDC *dc = printer.PrintDialog(wxGetApp().GetTopWindow());
    if (dc && dc->IsOk())
    {
      (*g_printData) = printer.GetPrintDialogData().GetPrintData();
      if (edit)
        SetEditorPrintSetup(edit,false,dc);
      printer.Print(wxGetApp().GetTopWindow(),printout,false);
    }
    delete dc;
    delete printout;
  }
  g_printData->SetOrientation(EFDoc->SeiteQuer ? wxLANDSCAPE : wxPORTRAIT);
}

wxPrintout* TEFView::CreatePrintout()
{
  wxWindow *win = Editor->FindFocus();
  TEFEquationsControl *eqct = wxDynamicCast(win,TEFEquationsControl);
  if (eqct)
    win = eqct->GetEditor();
  wxString name = EFDoc->GetUserReadableName();
  return new EFPrintOut(win,name);
}

bool TEFView::OnClose(bool deleteWindow)
{
  if (!GetDocument()->Close())
      return false;

  EFDoc = 0;
  if (deleteWindow)
  {
    Editor = 0;
    AkDiagW = 0;
    Activate(false);
    GetFrame()->Destroy();
  }
  SetFrame(0);
  return true;
}

void TEFView::OnGetFunktions(wxCommandEvent &)
{
  GetFunktionen(Funktion,true);
}

void TEFView::OnSetModified(wxCommandEvent &)
{
  EFDoc->SetIsDirty(true);
}

void TEFView::OnEditSetFocus(wxCommandEvent &event)
{
  TEFEdit *NeuEdit = wxDynamicCast(event.GetEventObject(),TEFEdit);
  AkEdit = NeuEdit;
}

bool TEFView::SolveEquations()
{
#ifdef EFMULTITHREAD
  TEFInteractiveSolverPanel *pnl;
  pnl = (TEFInteractiveSolverPanel *)GetFrame()->FindWindow(XRCID("pnlInteractiveSolver"));
  if (AkEdit == Editor)
    pnl->GetData().Solver->GetFehler() = SolverObj::NoEquationEdit;
  else
  if (pnl && AkEdit)
  {
    InteractiveSolverData &Data = pnl->GetData();
    GetFunktionen(Data.Fu,false,AkEdit);
    Data.Solver->StartWerte();
    return Data.Solver->InteractiveSolve(Data.Operand.c_str(),Data.Operator);
  }
#endif
  return false;
}

void TEFView::ShowInteractiveSolveResult(FunctionArrObj &Fu,SolverObj *Solver)
{
  if (AkEdit && (!Solver || !Solver->GetFehler()))
  {
    unsigned int Anf,End;
    AkEdit->GetSelection(Anf,End);
    Anf = AkEdit->GetLineFromPos(Anf);
    unsigned int End1 = End;
    End = AkEdit->GetLineFromPos(End1);
    if ((AkEdit->GetLineIndex(End) < End1) ||
        (AkEdit->GetLineIndex(End) == End1) && (End == Anf))
      if (End == AkEdit->GetNumLines()-1)
        End = AkEdit->GetTextLen();
      else
      {
        End++;
        End = AkEdit->GetLineIndex(End);
      }
    else
      End = AkEdit->GetLineIndex(End);
    Anf = AkEdit->GetLineIndex(Anf);
    AkEdit->SetSelection(Anf,End);

    MGPStr tmp;
    for (int i = 1;i <= Fu.GetMaxEintrag();i++)
    {
      Fu.SetAkEint(i);
      tmp.AddStr(Fu.GetAkEint().Namen);
      if (Fu.GetAkEint().Namen[0] && Fu.GetAkEint().Term[0])
        tmp.AddStr(wxT(" = "));
      tmp.AddStr(Fu.GetAkEint().Term);
      tmp.AddStr(wxT("\r\n"));
    }
    if (tmp.Length() > 0)
      AkEdit->Insert(tmp.c_str());
    int line = AkEdit->GetLineIndex(AkEdit->GetLineFromPos(-1));
    AkEdit->SetSelection(Anf,line);
  }
  else
  if (Solver)
  {
    wxMessageBox(TEFSolverDlg::FehlerMeldung(Solver->GetFehler()).c_str(),
                 wxT("Easy Funktion"),wxOK+wxICON_EXCLAMATION+wxCENTRE,
                 wxGetApp().GetTopWindow());
  }
}

void TEFView::OnWertBerechnen(wxCommandEvent &event)
{
  // Aktueller Editor wird uebergeben, wg. Abstuerzen mit AkEdit (?!)
  AkEdit = wxDynamicCast(event.GetEventObject(),TEFEdit);
  wxWindow *wnd = GetFrame()->FindWindow(XRCID("pnlEinzelwert"));
  if (AkEdit && wnd && wnd->IsShown())
  {
    TEFEinzelwertPanel *dlg = (TEFEinzelwertPanel *)wnd;
    dlg->SetEditor(AkEdit);
    wxCommandEvent evt(wxEVT_COMMAND_MENU_SELECTED,XRCID("ICM_WERTBERECHNEN"));
    dlg->GetEventHandler()->ProcessEvent(evt);
  }
  else // Ausgabe Funktionswert im "Tooltip"
  if (AkEdit)
  {
    ParserObj::FehlerRec Fehler;
    Cpx y(0,0);
    double x = 0;
    FunctionArrObj Fu;
    Fehler.ClearError();
    GetFunctionsIncludingModules(&Fu,true,AkEdit);
    if (event.GetInt() > Fu.GetMaxEintrag())
      return;
    Fu.SetAkEint(event.GetInt());
    EFDoc->GetParser().InitFunktions(&Fu);
    EFDoc->GetParser().FunktionsBerechnung(y,x,0,Fehler,0,true);
    EFDoc->GetParser().InitFunktions(Funktion);
    if (Fehler.Art == ParserObj::NoError)
    {
      wxString tmp;
      MGPStr tmp1;
      tmp = _("Function value (x = 0) : ");
      if (imag(y) != 0)
        tmp += _("Real :");
      EFDoc->GetParser().Ausgabe(tmp1,real(y),
                            wxGetApp().ZahlForm == ZahlFormatObj::IngForm ?
                            ZahlFormatObj::FitForm : ZahlFormatObj::ExpForm);
      tmp += wxString(tmp1.c_str());
      if (imag(y) != 0)
      {
        tmp += wxString(EFDoc->GetParser().GetLDV().ListSeparator.c_str());
        tmp += _(" Imag :");
        EFDoc->GetParser().Ausgabe(tmp1,imag(y),
                              wxGetApp().ZahlForm == ZahlFormatObj::IngForm ?
                              ZahlFormatObj::FitForm : ZahlFormatObj::ExpForm);
        tmp += wxString(tmp1.c_str());
      }
      tmp1.SetStr(tmp.c_str());
      AkEdit->ShowValueWindow(tmp1);
    }
  }
  if ((AkEdit == Editor) && AkEdit->IsModified())
  {
    UpdateAll(false);
//    Editor->SetFocus(); // stoert bei Gleichungsf. (entzieht manchmal Focus)
  }
}

void TEFView::GetEditorFont(TEFEdit *ed,FontType &Font)
{
  const wxFont &font = ed->GetFont();
  Font.Size = font.GetPointSize();
  Font.Font.SetStr(font.GetFaceName().c_str());
  Font.Style = (font.GetStyle() == wxITALIC ? FontType::Italic : 0) +
               (font.GetWeight() == wxBOLD ? FontType::Bold : 0) +
               (font.GetUnderlined() ? FontType::Underline : 0);
  wxColour clr = ed->GetForegroundColour();
  Font.Color = (unsigned long)(clr.Red() |
                              (clr.Green() << 8) |
                              (clr.Blue() << 16));
}

void TEFView::SetEditorFont(TEFEdit *ed,const FontType &Font)
{
  wxFont EditorFont(Font.Size,wxROMAN,
                    (Font.Style & FontType::Italic) ? wxITALIC : wxNORMAL,
                    (Font.Style & FontType::Bold) ? wxBOLD : wxNORMAL,
                    (Font.Style & FontType::Underline) > 0,
                    wxString(Font.Font.c_str()));
  ed->SetFont(EditorFont);
}

TEFEditSearch *TEFView::CreateEditor(wxFrame *parent)
{
    int width, height;
    parent->GetClientSize(&width, &height);

    TEFEditSearch *Editor = new TEFEditSearch(parent, ID_EDITOR, wxPoint(0, 0),
                                              wxSize(width, height),
                                              wxSUNKEN_BORDER |
                                              wxCLIP_CHILDREN |
                                              wxNO_FULL_REPAINT_ON_RESIZE |
                                              wxHSCROLL | wxVSCROLL |
                                              wxWANTS_CHARS,
                                              wxT("EF_Edit_Search"),
                                              wxGetApp().GetLDV());
    Editor->SetViewOption(wxGetApp().EditFlags & MyApp::EditScrollTrack,
                          TEFEdit::ViewScrollTrack,false);
//    Editor->SetExtraStyle(wxWS_EX_SCROLLBAR_ALWAYS);

    return Editor;
}

void TEFView::OnActivateView(bool activate, wxView *, wxView *)
{
  if (activate)
  {
    if (wxGetApp().zoombox && EFDoc)
    {
      wxString Zeile;
      Zeile.Printf(wxT("%i%%"),EFDoc->Zoom);
      int ind = wxGetApp().zoombox->FindString(Zeile);
      if (ind >= 0)
        wxGetApp().zoombox->SetSelection(ind);
    }
  }
}

bool TEFView::OnCreate(wxDocument *doc, long /*flags*/ )
{
  EFMDIChildFrame *frame = wxGetApp().CreateChildFrame(doc, this);

  Editor = CreateEditor(frame);
  frame->SetEditor(Editor);

  frame->Show(true);

  SetFrame(frame);

  g_printData = new wxPrintData;
  g_pageSetupData = new wxPageSetupDialogData;
  LabelEntryDlg = 0;
  BerDlg = 0;
  EFDoc = (TEFDocument *)doc;
  Funktion = EFDoc->Funktion;
  EFDoc->GetParser().InitFunktions(Funktion);
  Activate(true);
  AkDiagW = 0;
  LastCalcDiagW = 0;
  return true;
}

void TEFView::OnUpdate(wxView * /*sender*/,wxObject * /*hint*/)
{
  Activate(true); // Setzt korrekten Zoom in Zoombox
}

void TEFView::ParseFunktion(const MGPStr &Str,CpxObj &Funk,TEFEdit *edit)
{
  MGPStr Zeile(Str);
  if ((Zeile[0] <= ' ') && (Zeile[0] != 0))
  { // Kommentarzeile
    Funk.Namen.SetStr(Zeile);
    Funk.Term.SetStr(wxT(""));
  }
  else
  {
    ParserObj *Parser = &EFDoc->GetParser();
    int AkPos = Parser->IstGleichPos(Zeile.c_str());
    if (AkPos >= 0)
    { // '=' gefunden
      unsigned int TmpPos = AkPos+1;
      while ((TmpPos < Zeile.Length()) && (Zeile[TmpPos] == ' '))
        TmpPos++;
      Funk.Term.SetStr(&Zeile[TmpPos]);
      TmpPos = AkPos;
      while ((TmpPos > 0) && (Zeile[TmpPos-1] == ' '))
        TmpPos--;
      Zeile.ShortStr(TmpPos);
      if (!edit || (edit->GetEditorMode() == TEFEdit::FunktionStyle))
      {
        int APar = Parser->GetParamsAnfPos(Zeile.c_str());
        if (APar > 0)
        {
          Funk.ParamStr.SetStr(&Zeile[APar]);
          Zeile.ShortStr(APar);
        }
      }
      else
        Funk.ParamStr.SetStr(wxT(""));
      Funk.Namen.SetStr(Zeile);
    }
    else
    { // keine Bezeichnung
      Funk.Term.SetStr(Zeile);
      Funk.Namen.SetStr(wxT(""));
    }
  }
}

bool TEFView::GetSubmodulesFromDoc(TEFDocument *doc,FunctionArrObj *Fu,MGPStrSet &submodules,int &Count)
{
  Count++;

  // eventually break recursion
  if (Count > 10)
    return false;

  // Only load functions of submodules, as the functions of this module are obtained from editor
  if (Count > 1)
  {
    // Load functions of this module
    doc->Funktion->Begin();
    if (doc->Funktion->GetMaxEintrag() >= 0)
    do
    {
      Fu->NeuAkEint();
      // Only parse function if no name is already in place
      // (happens when file is loaded from old MCX format)
      if (doc->Funktion->GetAkEint().Namen.Length() == 0)
        ParseFunktion(doc->Funktion->GetAkEint().Term,Fu->GetAkEint(),0);
      else
      {
        Fu->GetAkEint().Term = doc->Funktion->GetAkEint().Term;
        Fu->GetAkEint().Namen = doc->Funktion->GetAkEint().Namen;
        Fu->GetAkEint().ParamStr = doc->Funktion->GetAkEint().ParamStr;
      }
    }
    while (doc->Funktion->NextEint());
  }

  // Go through all submodules of this document
  doc->m_submodules.Begin();
  if (doc->m_submodules.GetMaxEintrag() >= 0)
  do
  {
    // avoid recursion, only load document if not loaded previously
    if (submodules.find(doc->m_submodules.GetAkStr()) == submodules.end())
    {
      TEFDocument *tmp = new TEFDocument();

      submodules.insert(doc->m_submodules.GetAkStr());
      bool ret = tmp->LoadFile(doc->m_submodules.GetAkStr().c_str());

      if (ret)
        ret = GetSubmodulesFromDoc(tmp,Fu,submodules,Count);

      delete tmp;

      if (!ret)
        return false;
    }
  }
  while (doc->m_submodules.NextEint());

  Count--;
  return true;
}

bool TEFView::GetFunctionsIncludingModules(FunctionArrObj *Fu,bool AlleFkts,TEFEdit *edit)
{
  bool ret = GetFunktionen(Fu,AlleFkts,edit);

  // Remember currently active function
  int p = Fu->AkEintNummer();

  MGPStrSet Submodules;

  Submodules.insert(GetDocument()->GetFilename().EF_C_STR());
  int Count = 0;

  if (!GetSubmodulesFromDoc((TEFDocument *)GetDocument(),Fu,Submodules,Count))
    wxMessageBox(_("Recursive call of submodules !"));

  // Activate the currently active function
  Fu->SetAkEint(p);

  return ret;
}

bool TEFView::GetFunktionen(FunctionArrObj *Fu,bool AlleFkts,TEFEdit *edit)
{
  unsigned int AkFunk;
  unsigned int Anf,End;
  bool Gueltig;

  if (!edit)
    edit = Editor;
  edit->GetSelection(Anf,End);
  Anf = edit->GetLineFromPos(Anf);
  unsigned int End1 = End;
  End = edit->GetLineFromPos(End1);

  unsigned int p = 0;
  while ((edit->GetLineIndex(p) <= edit->GetLineIndex(-1)) &&
         (p < edit->GetNumLines()))
    p++;
  // nur 1 Zeile oder Cursor in letzter Zeile nicht am Anfang
  if ((End == Anf) || (edit->GetLineIndex(End) < End1))
    End++;
  Fu->ResetKette(); // bisherige Funktionen loeschen
  for (AkFunk = 0;AkFunk < edit->GetNumLines();AkFunk++)
  {
    Gueltig = (((AkFunk >= Anf) && (AkFunk < End)) || AlleFkts);
    if (Gueltig && !((AkFunk+1 == edit->GetNumLines()) &&
        (edit->GetLineLength(AkFunk) == 0)))
    {
      MGPStr Zeile;
      Zeile = edit->GetLine(AkFunk);
      Fu->NeuAkEint();
      Fu->GetAkEint().Status = (edit->IsLineModified(AkFunk+1) ?
                           CpxObj::Modified : 0);
      ParseFunktion(Zeile,Fu->GetAkEint(),edit);
      if ((edit == Editor) &&
          (Fu->GetAkEint().Namen.Length() > 0) &&
          (Fu->GetAkEint().Term.Length() > 0) &&
          (Fu->GetAkEint().Namen[0] != '['))
      { // FIXME: indirectly modifies caret-position!!
        Fu->GetAkEint().Namen.InsChar(0,'[');
        Fu->GetAkEint().Namen.AddChar(']');
      }
    }
  }
  Fu->SetAkEint(p);
  if (Fu == Funktion) // Nur initialisieren wenn FunktionsFeld des Dokuments
    EFDoc->GetParser().InitFunktions(Funktion);
  return true;
}

void TEFView::OnDraw(wxDC *dc)
{
  if (Editor)
  {
    Editor->SetPrintDocSize(*dc);
    Editor->PaintPage(*dc,1);
    Editor->SetScreenDocSize();
  }
}

void TEFView::ClearModify()
{
  if (!Editor)
    return;
  Editor->ClearModify();
  wxWindowListNode *Child = Editor->GetChildren().GetFirst();
  while (Child)
  {
    TEFEquationsControl *edit = wxDynamicCast((wxWindow *)Child->GetData(),
                                              TEFEquationsControl);
    if (edit)
      edit->ClearModify();
    Child = Child->GetNext();
  }
}

void TEFView::VnCommit(DiagDataPtArrObj &DiagDataPts)
{ // Daten in das Dokument zurueckschreiben
  wxClientDC dc(Editor);
  wxSize size = dc.GetPPI();
  GetFunktionen(Funktion,true);
  EFDoc->ScrollLine = Editor->GetFirstVisibleLine();
  Editor->GetSelection(EFDoc->EditSelAnf,EFDoc->EditSelEnd);
  wxSize s = Editor->GetScrollSize();
  EFDoc->ScrollPosX = int(s.x*96.0/size.x);
  EFDoc->ScrollPosY = int(s.y*96.0/size.y);
  BildschOptObj &BO = EFDoc->BildschOpt;
  BO.SetBildschKlammer(Editor->GetViewOption(TEFEdit::ViewKlammerDarst));
  BO.SetBildschBruch(Editor->GetViewOption(TEFEdit::ViewBruchDarst));
  BO.SetHideDiags(Editor->GetViewOption(TEFEdit::ViewHideDiags));
  BO.SetMarkLabels(Editor->GetViewOption(TEFEdit::ViewMarkLabels));
  Editor->ClearModify();
  wxWindowListNode *child = Editor->GetChildren().GetFirst();
  while (child)
  {
    wxPictureControl *tmp;
    tmp = wxDynamicCast(child->GetData(),TEFBitmapControl);
#ifdef __WXMSW__
    if (!tmp)
      tmp = wxDynamicCast(child->GetData(),TEFMetafileControl);
#endif
    if (tmp)
    { // Bild
      EFDoc->PictureArr->NeuAkEint();
      wxRect rect = tmp->GetRect();
      EFDoc->PictureArr->AkPicture()->TypeStr.SetStr(wxT("XPM"));
      wxImage img;
#ifdef __WXMSW__
      if (!_tcscmp(tmp->GetPictureType(),wxT("EnhMetaFile")))
      {
        if (!(wxGetApp().Options & MyApp::keySaveEMFAsBitmap))
        {
          EFDoc->PictureArr->AkPicture()->TypeStr.SetStr(tmp->GetPictureType());
          EFDoc->PictureArr->AkPicture()->Size = tmp->GetPictureSize();
          EFDoc->PictureArr->AkPicture()->Data =
            new char[EFDoc->PictureArr->AkPicture()->Size];
          tmp->GetPictureData(EFDoc->PictureArr->AkPicture()->Data,
                              EFDoc->PictureArr->AkPicture()->Size);
        }
        else
        {
          wxMemoryDC dc;
          wxMetafileControl *mc = wxDynamicCast(child->GetData(),
                                                wxMetafileControl);

          wxASSERT(mc);

          wxBitmap bmp(mc->GetMetafile().GetWidth(),
                       mc->GetMetafile().GetHeight());
          dc.SelectObject(bmp);
          wxBrush OrgBr = dc.GetBrush();
          dc.SetBrush(*wxWHITE_BRUSH);
          dc.Clear();
          ((wxEnhMetaFile)mc->GetMetafile()).Play(&dc);
          dc.SelectObject(wxNullBitmap);
          dc.SetBrush(OrgBr);
          img = bmp.ConvertToImage();
        }
      }
      else
#endif
      {
        wxBitmapControl *bmc = wxDynamicCast(child->GetData(),wxBitmapControl);

        wxASSERT(bmc);

        img = bmc->GetBitmap().ConvertToImage();
      }
      if (_tcscmp(tmp->GetPictureType(),wxT("EnhMetaFile")) ||
          (wxGetApp().Options & MyApp::keySaveEMFAsBitmap))
      {
        wxMemoryOutputStream str;
        img.SaveFile(str,wxBITMAP_TYPE_XPM);
        EFDoc->PictureArr->AkPicture()->Data = new char[str.GetSize()];
        str.CopyTo(EFDoc->PictureArr->AkPicture()->Data,str.GetSize());
        EFDoc->PictureArr->AkPicture()->Size = str.GetSize();
      }
      EFDoc->PictureArr->AkPicture()->WinXPos = int(rect.x*96.0/size.x);
      EFDoc->PictureArr->AkPicture()->WinYPos = int(rect.y*96.0/size.y);
      EFDoc->PictureArr->AkPicture()->WinHeight = int(rect.height*96.0/size.x);
      EFDoc->PictureArr->AkPicture()->WinWidth = int(rect.width*96.0/size.y);
    }
    TEFGraphControl *gr;
    gr = wxDynamicCast(child->GetData(),TEFGraphControl);
    if (gr)
    {
      DiagDataPts.NeuAkEint();
      DiagDataPts.GetAkEint() = gr->GetDiagData();
      wxRect rect = gr->GetRect();
      DiagDataPts.GetAkEint()->WinXPos = int(rect.x*96.0/size.x);
      DiagDataPts.GetAkEint()->WinYPos = int(rect.y*96.0/size.y);
      DiagDataPts.GetAkEint()->WinHeight = int(rect.height*96.0/size.x);
      DiagDataPts.GetAkEint()->WinWidth = int(rect.width*96.0/size.y);
    }
    TEFEquationsControl *eq;
    eq = wxDynamicCast(child->GetData(),TEFEquationsControl);
    if (eq)
    {
      wxRect rect = eq->GetRect();
      EFDoc->EquationsArr->NeuAkEint();
      EquationsObj *tmp = EFDoc->EquationsArr->AkEquations();
      TEFEdit *ed = eq->GetEditor();
      BildschOptObj &BO = tmp->BildschOpt;
      BO.SetMarkLabels(ed->GetViewOption(TEFEdit::ViewMarkLabels));
      BO.SetBildschBruch(ed->GetViewOption(TEFEdit::ViewBruchDarst));
      BO.SetBildschKlammer(ed->GetViewOption(TEFEdit::ViewKlammerDarst));
      BO.SetMarkLabels(ed->GetViewOption(TEFEdit::ViewMarkLabels));
      eq->GetData(tmp->Equations);
      tmp->WinXPos = int(rect.x*96.0/size.x);
      tmp->WinYPos = int(rect.y*96.0/size.y);
      tmp->WinHeight = int(rect.height*96.0/size.x);
      tmp->WinWidth = int(rect.width*96.0/size.y);
      GetEditorFont(ed,tmp->Font);
      ed->GetSelection(tmp->SelAnf,tmp->SelEnd);
      wxSize s = ed->GetScrollSize();
      tmp->ScrollPosX = int(s.x*96.0/size.x);
      tmp->ScrollPosY = ed->GetFirstVisibleLine();
    }
    child = child->GetNext();
  }
  const wxFont &font = Editor->GetFont(); // Aktuelle Schriftgroesse laden
  EditFont.Size = font.GetPointSize();
  EFDoc->Editor = EditFont;
  EFDoc->SeiteQuer = Editor->GetSeitenAusrichtung();

  EFDoc->m_OpenPanels.ResetKette(false);
  child = GetFrame()->GetChildren().GetFirst();
  while (child)
  {
    wxWindow *tmp;
    tmp = wxDynamicCast(child->GetData(),TEFEinzelwertPanel);
    if (tmp)
    {
      EFDoc->m_OpenPanels.NeuAkEint();
      EFDoc->m_OpenPanels.GetAkEint() = EFDaten::SingleValuePanelOpen + (tmp->IsShown() ? EFDaten::PanelVisible : 0);
    }

    tmp = wxDynamicCast(child->GetData(),TEFFunktionsPanel);
    if (tmp)
    {
      EFDoc->m_OpenPanels.NeuAkEint();
      EFDoc->m_OpenPanels.GetAkEint() = EFDaten::FunctionsPanelOpen + (tmp->IsShown() ? EFDaten::PanelVisible : 0);
    }

    tmp = wxDynamicCast(child->GetData(),TEFInteractiveSolverPanel);
    if (tmp)
    {
      EFDoc->m_OpenPanels.NeuAkEint();
      EFDoc->m_OpenPanels.GetAkEint() = EFDaten::InteractiveSolverPanelOpen + (tmp->IsShown() ? EFDaten::PanelVisible : 0);
    }

    tmp = wxDynamicCast(child->GetData(),TEFSubmodulesPanel);
    if (tmp)
    {
      EFDoc->m_OpenPanels.NeuAkEint();
      EFDoc->m_OpenPanels.GetAkEint() = EFDaten::SubmodulesPanelOpen + (tmp->IsShown() ? EFDaten::PanelVisible : 0);
    }

    child = child->GetNext();
  }
}

void TEFView::SetEditorFunk()
{
  int p = Funktion->AkEintNummer();
  MGPStr tmp;
  Funktion->Begin();
  while (Funktion->NextEint())
  {
    tmp.AddStr((*Funktion)->Namen);
    tmp.AddStr((*Funktion)->ParamStr);
    if (((*Funktion)->Term.Length() > 0) &&
        ((*Funktion)->Namen.Length() > 0))
      tmp.AddStr(wxT(" = "));
    tmp.AddStr((*Funktion)->Term);
    tmp.AddStr(wxT("\r\n"));
  }
  Editor->Insert(tmp.c_str(),false,false,false); // Zeilen einfuegen, aber nicht rollen
  Funktion->SetAkEint(p);
};

void TEFView::OnSetAkDiag(wxCommandEvent &event)
{
  TEFGraphControl *GG = wxDynamicCast(event.GetEventObject(),TEFGraphControl);
  SetAkDiag(GG);
}

void TEFView::SetAkDiag(TEFGraphControl *GG)
{ // sucht GG in Editor-Child Liste und aktiviert es
  AkDiagW = GG;
  wxWindowListNode *child = Editor->GetChildren().GetFirst();
  while (child)
  {
    TEFGraphControl *tmp = wxDynamicCast(child->GetData(),TEFGraphControl);
    if (tmp)
    {
      tmp->SetActive(child->GetData() == AkDiagW);
      if ((child->GetData() == AkDiagW) && (tmp->FindFocus() != AkDiagW))
        tmp->SetFocus();
    }
    child = child->GetNext();
  }
}

void TEFView::NeuAkDiag(DiagDataArrObj *DDA,bool usescrollposy)
{
  wxClientDC dc(Editor);
  wxSize res = dc.GetPPI();
  int y = DDA->AkDiagData()->WinYPos;
  if (usescrollposy)
    y += EFDoc->ScrollPosY;
  wxPoint pos(int(DDA->AkDiagData()->WinXPos*float(res.x)/96),
              int(y*float(res.y)/96));
  wxSize size(int(DDA->AkDiagData()->WinWidth*float(res.x)/96),
              int(DDA->AkDiagData()->WinHeight*float(res.y)/96));
  AkDiagW = new TEFGraphControl();
  AkDiagW->Create(Editor,0,DDA->AkDiagData(),EFDoc->GetParser().GetLDV(),pos,size,
#if wxCHECK_VERSION(2,5,1)
                  wxFULL_REPAINT_ON_RESIZE |
#endif
                  wxCLIP_SIBLINGS);
  AkDiagW->SetCurrentZoom(Editor->GetInternalZoom());
  SetAkDiag(AkDiagW);
  RenumberChildWindows();
}

bool TEFView::LoadData(bool NewDiag)
{
  MGPStr Zeile;

  Editor->Enable(false);
  Editor->SetLabelFontStyle(wxGetApp().LFontStyle);
  Editor->SetKommentFontStyle(wxGetApp().KFontStyle);
  Editor->SetLabelFontColour(wxGetApp().LFontColour);
  Editor->SetKommentFontColour(wxGetApp().KFontColour);
  Editor->ClearEditor();
  SetEditorPrintSetup(Editor,true);
  BildschOptObj &BO = EFDoc->BildschOpt;
  Editor->SetViewOption(BO.GetBildschBruch(),TEFEdit::ViewBruchDarst);
  Editor->SetViewOption(BO.GetBildschKlammer(),TEFEdit::ViewKlammerDarst);
  EditFont = EFDoc->Editor;
  SetEditorFont(Editor,EditFont);
  Editor->SetZoom(0.01*EFDoc->Zoom);
  Editor->SetViewOption(BO.GetMarkLabels(),TEFEdit::ViewMarkLabels);
  SetEditorFunk();
  // Pruefen ob aktuelle Selektion nicht gespeichert (Version < 4.21)
  if (EFDoc->EditSelAnf > EFDoc->EditSelEnd)
    Editor->SetSelection(Editor->GetLineIndex(Funktion->AkEintNummer()-1),
                         Editor->GetLineIndex(Funktion->AkEintNummer()-1),
                         false);
  else // Aktuelle Selektion gespeichert
    Editor->SetSelection(EFDoc->EditSelAnf,EFDoc->EditSelEnd,false);

  wxSize size1 = Editor->GetScrollSize();
  if (EFDoc->ScrollPosY && size1.y || size1.x) // wenn y-Scrollpos, auf 0 rollen
    Editor->Scroll(-size1.x,-Editor->GetFirstVisibleLine());

  wxClientDC dc(Editor);
  wxSize size = dc.GetPPI();
  if (!EFDoc->ScrollPosY)
    Editor->Scroll(int(EFDoc->ScrollPosX*float(size.x)/96),
                   EFDoc->ScrollLine-Editor->GetFirstVisibleLine());

  DiagDataArrObj *DDA = EFDoc->DiagDataArr;
  DDA->Begin();
  while (DDA->GetMaxEintrag() >= 0)
  {
    NeuAkDiag(DDA,true);
    DDA->LoeschAkEint();
    DDA->Begin();
  }
  if (NewDiag)
    CreateAkDiag();
  PictureArrObj *pic = EFDoc->PictureArr;
  pic->Begin();
  if (pic->GetMaxEintrag() >= 0)
    do
    {
      if (pic->AkPicture()->Data)
      {
        int y = pic->AkPicture()->WinYPos+EFDoc->ScrollPosY;
        wxPoint pos = wxPoint(int(pic->AkPicture()->WinXPos*float(size.x)/96),
                              int(y*float(size.y)/96));
        wxSize sz =  wxSize(int(pic->AkPicture()->WinWidth*float(size.x)/96),
                            int(pic->AkPicture()->WinHeight*float(size.y)/96));
        if (_tcscmp(pic->AkPicture()->TypeStr.c_str(),wxT("XPM")) == 0)
        {
          wxMemoryInputStream str(pic->AkPicture()->Data,pic->AkPicture()->Size);
          wxImage img;
          img.LoadFile(str,wxBITMAP_TYPE_XPM);
          TEFBitmapControl *tmp = new TEFBitmapControl();
          tmp->Create(Editor,0,wxBitmap(img),pos,sz,
#if wxCHECK_VERSION(2,5,1)
                      wxFULL_REPAINT_ON_RESIZE |
#endif
                      wxCLIP_SIBLINGS);
          delete pic->AkPicture()->Data;
          pic->AkPicture()->Data = 0;
          pic->AkPicture()->Size = 0;
          tmp->SetBackgroundColour(WINDOWCOLOUR);
#if wxCHECK_VERSION(2,5,0)
          tmp->ClearBackground();
#else
          tmp->Clear();
#endif
          tmp->SetCurrentZoom(Editor->GetInternalZoom());
        }
#ifdef __WXMSW__
        else
        if (_tcscmp(pic->AkPicture()->TypeStr.c_str(),wxT("Dib")) == 0)
        {
          TEFBitmapControl *tmp = new TEFBitmapControl();
          tmp->Create(Editor,0,pic->AkPicture()->Data,
                      pic->AkPicture()->Size,pos,sz,
#if wxCHECK_VERSION(2,5,1)
                      wxFULL_REPAINT_ON_RESIZE |
#endif
                      wxCLIP_SIBLINGS);
          delete pic->AkPicture()->Data;
          pic->AkPicture()->Data = 0;
          pic->AkPicture()->Size = 0;
          tmp->SetBackgroundColour(WINDOWCOLOUR);
#if wxCHECK_VERSION(2,5,0)
          tmp->ClearBackground();
#else
          tmp->Clear();
#endif
          tmp->SetCurrentZoom(Editor->GetInternalZoom());
        }
        else
        {
          TEFMetafileControl *tmp = new TEFMetafileControl();
          tmp->Create(Editor,0,pic->AkPicture()->Data,
                      pic->AkPicture()->Size,pos,sz,
#if wxCHECK_VERSION(2,5,1)
                      wxFULL_REPAINT_ON_RESIZE |
#endif                
                      wxCLIP_SIBLINGS);
          delete pic->AkPicture()->Data;
          pic->AkPicture()->Data = 0;
          pic->AkPicture()->Size = 0;
          tmp->SetBackgroundColour(WINDOWCOLOUR);
#if wxCHECK_VERSION(2,5,0)
          tmp->ClearBackground();
#else
          tmp->Clear();
#endif
          tmp->SetCurrentZoom(Editor->GetInternalZoom());
        }
#endif
      }
    }
    while (pic->NextEint());
  EquationsArrObj *eq = EFDoc->EquationsArr;
  eq->Begin();
  if (eq->GetMaxEintrag() >= 0)
    do
    {
      int y = eq->AkEquations()->WinYPos+EFDoc->ScrollPosY;
      wxPoint pos = wxPoint(int(eq->AkEquations()->WinXPos*float(size.x)/96),
                            int(y*float(size.y)/96));
      wxSize sz =  wxSize(int(eq->AkEquations()->WinWidth*float(size.x)/96),
                          int(eq->AkEquations()->WinHeight*float(size.y)/96));
      wxSize SelSz = wxSize(eq->AkEquations()->SelAnf,eq->AkEquations()->SelEnd);
      wxSize ScPos = wxSize(int(eq->AkEquations()->ScrollPosX*float(size.x)/96),
                            eq->AkEquations()->ScrollPosY);
      NeuEquations(pos,sz,eq->AkEquations()->Equations.c_str(),
                   &eq->AkEquations()->Font,&eq->AkEquations()->BildschOpt,
                   &SelSz,&ScPos);
    }
    while (eq->NextEint());
  RenumberChildWindows();
  if (EFDoc->ScrollPosY)
    Editor->Scroll(int(EFDoc->ScrollPosX*float(size.x)/96),
                   EFDoc->ScrollLine-Editor->GetFirstVisibleLine());
  Editor->SetViewOption(EFDoc->BildschOpt.GetHideDiags(),
                        TEFEdit::ViewHideDiags);
  Editor->ClearLineModified();
  Editor->Enable(true);
  Editor->SetFocus();

  EFDoc->m_OpenPanels.Begin();
  bool funcpanelshown = false;
  if (EFDoc->m_OpenPanels.GetMaxEintrag() >= 0)
  {
    wxCommandEvent evt(wxEVT_COMMAND_MENU_SELECTED,0);
    do
    {
      if (EFDoc->m_OpenPanels.GetAkEint() & EFDaten::SingleValuePanelOpen)
      {
        OnEinzWert(evt);
        // Hide panel if not visible flag set
        if (!(EFDoc->m_OpenPanels.GetAkEint() & EFDaten::PanelVisible))
          OnEinzWert(evt);
      }
      else
      if (EFDoc->m_OpenPanels.GetAkEint() & EFDaten::FunctionsPanelOpen)
      {
        OnFunktionen(evt);
        funcpanelshown = true;
        // Hide panel if not visible flag set
        if (!(EFDoc->m_OpenPanels.GetAkEint() & EFDaten::PanelVisible))
          OnFunktionen(evt);
      }
      else
      if (EFDoc->m_OpenPanels.GetAkEint() & EFDaten::InteractiveSolverPanelOpen)
      {
        OnInteractiveSolver(evt);
        // Hide panel if not visible flag set
        if (!(EFDoc->m_OpenPanels.GetAkEint() & EFDaten::PanelVisible))
          OnInteractiveSolver(evt);
      }
      else
      if (EFDoc->m_OpenPanels.GetAkEint() & EFDaten::SubmodulesPanelOpen)
      {
        OnSubmodules(evt);
        // Hide panel if not visible flag set
        if (!(EFDoc->m_OpenPanels.GetAkEint() & EFDaten::PanelVisible))
          OnSubmodules(evt);
      }

    }
    while (EFDoc->m_OpenPanels.NextEint());
  }

  if ((wxGetApp().Options & MyApp::keyFuncSelVisible) && !funcpanelshown)
  {
    wxCommandEvent evt(wxEVT_COMMAND_MENU_SELECTED,XRCID("ICM_SELECTFUNCTION"));
    AddPendingEvent(evt);
  }

  wxCommandEvent evt(wxEVT_COMMAND_MENU_SELECTED,XRCID("ICM_CLEARMODIFIED"));
  AddPendingEvent(evt);

  return true;
}

void TEFView::NeuEquations(const wxPoint &pos,const wxSize &size,
                           const _TCHAR *data,const FontType *Font,
                           const BildschOptObj *BOpt,
                           const wxSize *Selection,
                           const wxSize *ScrollPos)
{
  TEFEquationsControl *tmp = new TEFEquationsControl();
  tmp->Create(Editor,0,pos,size,
              wxCLIP_SIBLINGS | wxCLIP_CHILDREN |
              wxNO_FULL_REPAINT_ON_RESIZE | wxWANTS_CHARS,
              wxGetApp().GetLDV(),
              wxGetApp().EditFlags & MyApp::EditScrollTrack);
  tmp->SetCurrentZoom(Editor->GetInternalZoom());
  TEFEdit *Ed = tmp->GetEditor();
  Ed->SetZoom(0.01*EFDoc->Zoom);
  if (Font)
    SetEditorFont(Ed,*Font);
  Ed->SetLabelFontStyle(wxGetApp().LFontStyle);
  Ed->SetKommentFontStyle(wxGetApp().KFontStyle);
  Ed->SetLabelFontColour(wxGetApp().LFontColour);
  Ed->SetKommentFontColour(wxGetApp().KFontColour);
  if (BOpt)
  {
    Ed->SetViewOption(BOpt->GetBildschBruch(),TEFEdit::ViewBruchDarst);
    Ed->SetViewOption(BOpt->GetBildschKlammer(),TEFEdit::ViewKlammerDarst);
    Ed->SetViewOption(BOpt->GetMarkLabels(),TEFEdit::ViewMarkLabels);
    Ed->SetViewOption(true,TEFEdit::ViewHideDiags);
  }
  if (data)
    tmp->SetData(MGPStr(data));
  if (Selection)
    Ed->SetSelection(Selection->x,Selection->y,false);
  if (ScrollPos)
  {
    wxSize size = tmp->GetEditor()->GetScrollSize();
    Ed->Scroll(ScrollPos->x-size.x,
               ScrollPos->y-Ed->GetFirstVisibleLine());
  }
  SetEditorPrintSetup(Ed,true);
  tmp->ClearModify();
  RenumberChildWindows();
}

void TEFView::OnNeuEquations(wxCommandEvent &)
{
  NeuEquations(wxPoint(100,100),wxSize(100,100));
  EFDoc->SetIsDirty(true);
}

void TEFView::OnNeuAkDiag(wxCommandEvent &)
{
  CreateAkDiag();
}

void TEFView::CreateAkDiag()
{
  DiagDataArrObj *DDA = EFDoc->DiagDataArr;
  DDA->NeuAkEint();
  DDA->GetAkEint().DarstData[0].Mode |= (wxGetApp().DiagFlags & MyApp::DiagHasFrame ?
                                         GraphBaseObj::DiaRahmen : 0);
  DDA->GetAkEint().DarstData[1].Mode |= (wxGetApp().DiagFlags & MyApp::DiagHasFrame ?
                                         GraphBaseObj::DiaRahmen : 0);
  DDA->GetAkEint().DarstData[0].Mode |= (wxGetApp().DiagFlags & MyApp::DiagEndLabels ?
                                         GraphBaseObj::DiaEndLabels : 0);
  DDA->GetAkEint().DarstData[1].Mode |= (wxGetApp().DiagFlags & MyApp::DiagEndLabels ?
                                         GraphBaseObj::DiaEndLabels : 0);
  DDA->GetAkEint().DarstData[0].Mode |= (wxGetApp().DiagFlags & MyApp::DiagXValuesVert ?
                                         GraphBaseObj::DiaXVertikal : 0);
  DDA->GetAkEint().DarstData[1].Mode |= (wxGetApp().DiagFlags & MyApp::DiagXValuesVert ?
                                         GraphBaseObj::DiaXVertikal : 0);
  DDA->GetAkEint().m_Colour = wxGetApp().DiagColour;
  NeuAkDiag(DDA);
  EFDoc->SetIsDirty(true);
}

void TEFView::PasteBild()
{ // fuegt ein Bild aus der Zwischenablage ein
#ifdef __WXMSW__
  if (wxTheClipboard->IsSupported(wxDF_ENHMETAFILE))
  {
    wxEnhMetaFileDataObject cdo;
    wxTheClipboard->GetData(cdo);
    WXHANDLE cpy;
    cdo.GetDataHere(wxDF_ENHMETAFILE,&cpy);
    TEFMetafileControl *AkBild = new TEFMetafileControl();
    AkBild->Create(Editor,0,cpy,wxPoint(10,10),wxSize(10,10),
#if wxCHECK_VERSION(2,5,1)
                   wxFULL_REPAINT_ON_RESIZE |
#endif
                   wxCLIP_SIBLINGS);
    AkBild->SetBackgroundColour(WINDOWCOLOUR);
#if wxCHECK_VERSION(2,5,0)
    AkBild->ClearBackground();
#else
    AkBild->Clear();
#endif
    AkBild->SetFocus();
  }
  else
#endif
  if (wxTheClipboard->IsSupported(wxDF_BITMAP))
  {
    wxBitmapDataObject cdo;
    wxTheClipboard->GetData(cdo);
    wxSize sz(cdo.GetBitmap().GetWidth(),cdo.GetBitmap().GetHeight());
    TEFBitmapControl *AkBild = new TEFBitmapControl();
    AkBild->Create(Editor,0,cdo.GetBitmap(),wxPoint(10,10),sz,
#if wxCHECK_VERSION(2,5,1)
                   wxFULL_REPAINT_ON_RESIZE |
#endif
                   wxCLIP_SIBLINGS);
    AkBild->SetBackgroundColour(WINDOWCOLOUR);
#if wxCHECK_VERSION(2,5,0)
    AkBild->ClearBackground();
#else
    AkBild->Clear();
#endif
    AkBild->SetFocus();
  }
  RenumberChildWindows();
}

void TEFView::OnEditPaste(wxCommandEvent &event)
{
#ifdef __WXGTK__
  if (wxTheClipboard->IsOpened())
    return;
#endif
  if (AkDiagW && (AkDiagW->FindFocus() == AkDiagW))
    AkDiagW->OnEditPaste(event);
  else
  {
    wxString choices[2];
    int count = 0;
#ifdef __WXMSW__
    if (wxTheClipboard->IsSupported(wxDF_METAFILE) ||
        wxTheClipboard->IsSupported(wxDF_BITMAP))
#else
    if (wxTheClipboard->IsSupported(wxDF_BITMAP))
#endif
    {
      choices[count] = _("Graphics");
      count++;
    }
    if (wxTheClipboard->IsSupported(wxDF_TEXT))
    {
      choices[count] = wxT("Text");
      count++;
    }
    int sel = -1;
    if ((count > 1) &&
        !wxTheClipboard->IsSupported(wxGetApp().editformat))
    {
      wxSingleChoiceDialog dlg(Editor,_("Select clipboard format to insert"),
                               _("Insert data"),count,choices);
      if (dlg.ShowModal() == wxID_OK)
        sel = dlg.GetSelection();

    }

    wxUpdateUIEvent ce;
    Editor->CeEditPaste(ce); // pruefen ob Editor-Format vorhanden
    if (ce.GetEnabled() && ((sel == -1) || !_tcscmp(choices[sel],wxT("Text"))))
    {
      TEFEdit *edit = wxDynamicCast(Editor->FindFocus(),TEFEdit);
      if (edit)
        edit->OnEditPaste(ce);
    }
    else // nur wenn nicht schon Editor eingefuegt
    if (wxTheClipboard->Open())
    {
      if (wxTheClipboard->IsSupported(wxGetApp().diagformat))
      { // Diagramm im Clipboard
        EFDoc->DiagDataArr->NeuAkEint();
        NeuAkDiag(EFDoc->DiagDataArr);
        wxTheClipboard->Close();
        AkDiagW->OnEditPaste(event);
        wxTheClipboard->Open();
        EFDoc->SetIsDirty(true);
      }
      else
#ifdef __WXMSW__
      if (wxTheClipboard->IsSupported(wxDF_METAFILE) ||
          wxTheClipboard->IsSupported(wxDF_BITMAP))
#else
      if (wxTheClipboard->IsSupported(wxDF_BITMAP))
#endif
      { // wenn Bild im Clipboard
        PasteBild();
      }
      wxTheClipboard->Close();
    }
  }
}

int TEFView::GetDiagCount()
{
  int count = 0;
  wxWindowListNode *Child = Editor->GetChildren().GetFirst();
  while (Child)
  {
    TEFGraphControl *gc = wxDynamicCast((wxWindow *)Child->GetData(),
                                        TEFGraphControl);
    if (gc)
      count++;
    Child = Child->GetNext();
  }
  return count;
}

void TEFView::CeDelDiag(wxUpdateUIEvent &ce)
{
  ce.Enable(GetDiagCount() > 0);
}

void TEFView::CeEditPaste(wxUpdateUIEvent& ce)
{
  ce.Enable(false);
#ifdef __WXGTK__
  if (wxTheClipboard->IsOpened())
    return;
#endif
  if (AkDiagW && (AkDiagW->FindFocus() == AkDiagW))
    AkDiagW->CeEditPaste(ce);
  else
  {
    Editor->CeEditPaste(ce);
    wxTheClipboard->Open();
    ce.Enable(ce.GetEnabled() ||
#ifdef __WXMSW__
               wxTheClipboard->IsSupported(wxDF_METAFILE) ||
#endif
               wxTheClipboard->IsSupported(wxDF_BITMAP) ||
               wxTheClipboard->IsSupported(wxGetApp().diagformat));
    wxTheClipboard->Close();
  }
}

void TEFView::OnDelWindow(wxCommandEvent &event)
{
  wxWindow *wnd = wxDynamicCast(event.GetEventObject(),wxWindow);
  if (wnd &&
      (Editor->GetChildren().IndexOf(wnd) != wxNOT_FOUND))
  {
    int test = wxYES;
    if ((wxGetApp().Options & MyApp::keyWarnDelete) &&
        !MyMacroRecorder::Get()->IsPlaying())
      test = wxMessageBox(_("Delete active window ?\n\nThis command cannot be undone !"),
                          wxT("Easy Funktion"),
                          wxYES_NO+wxICON_QUESTION+wxCENTRE,
                          wxGetApp().GetTopWindow());
    if (wxYES == test)
    {
      wnd->Destroy();
      if (wnd == AkDiagW)
         AkDiagW = 0;
      if (wnd == AkEdit)
         AkEdit = 0;
      EFDoc->SetIsDirty(true);
    }
  }
}

void TEFView::OnZoom(wxCommandEvent &)
{
  EFDoc->Zoom = wxGetApp().Zoom;
  Editor->SetZoom(0.01*EFDoc->Zoom);
  wxWindowListNode *Child = Editor->GetChildren().GetFirst();
  while (Child)
  {
    TEFEquationsControl *edit = wxDynamicCast(Child->GetData(),
                                              TEFEquationsControl);
    if (edit)
      edit->SetZoom(0.01*EFDoc->Zoom);
    Child = Child->GetNext();
  }
  Editor->SetFocus();
  EFDoc->SetIsDirty(true);
}

void TEFView::OnPageBreak(wxCommandEvent &)
{
  Editor->InsertPageBreak();
}

void TEFView::OnEditFont(wxCommandEvent &)
{
  TEFEdit *edit = wxDynamicCast(Editor->FindFocus(),TEFEdit);
  if (!edit)
    return;

  FontType Font;
  GetEditorFont(edit,Font);
  TEFHilfeDlg::GetFont(Editor,Font,_("Function editor"));
  SetEditorFont(edit,Font);
  if (edit == Editor)
    EFDoc->Editor = Font;
  EFDoc->SetIsDirty(true);
  if (MyMacroRecorder::Get()->IsRecording())
  {
    MyMacroRecorder::Get()->AddEvent(wxT("CMD_EDITFONT"),edit->GetId());
    EFXmlNode *node = MyMacroRecorder::Get()->GetAkNode();
    EFDoc->XmlWriteEditFont(node,Font);
  }  
}

void TEFView::OnUpdateAll(wxCommandEvent &)
{
  UpdateAll(true);
  Editor->SetFocus();
}

// ermittelt den Namen der aktuellen Funktion ohne Klammern
void TEFView::AkFuNameOhneKlammern(MGPStr &nm)
{
  nm.AddStr((*Funktion)->Namen);
  if ((nm[0] == '[') && (nm[nm.Length()-1] == ']'))
  { // evtl. Klammern entfernen
    nm.DelNChar(0,1);
    nm.ShortStr(nm.Length()-1);
  }
}

void TEFView::UpdateAll(bool UpdateAkDiag)
{
  bool Erg;
  int AkFunk,i;
  MGPStr FuStr;

  if (AkDiagW || !UpdateAkDiag)
  {
    TEFGraphControl *OrgDiag = AkDiagW;
    AkFunk = Funktion->AkEintNummer();
    wxWindowListNode *child = (wxWindowListNode *)1; // dummy, wg. Schleifenbedingung
    if (!UpdateAkDiag)
      child = Editor->GetChildren().GetFirst();
    bool modifiedany = false;
    GetFunctionsIncludingModules(Funktion,true); // Funktionen neu laden
    // Alle in allen Diagrammen berechnete Funktionen
    vector<bool> GesBerFus(Funktion->GetMaxEintrag()+1);
    while (child)
    { // alle Diagramme
      TEFGraphControl *tmp = AkDiagW;
      if (!UpdateAkDiag)
        tmp = wxDynamicCast(child->GetData(),TEFGraphControl);
      if (tmp && (tmp->GetDiagData()->AutoUpdate || UpdateAkDiag))
      { // wenn Fenster Diagramm
        i = 0;
        Erg = true;
        GraphManagerObj *YDaten = &tmp->GetDiagData()->YDaten;
        MaTyp *Ma = &tmp->GetDiagData()->Ma;
        // Aufgerufene Funktionen
        vector<bool> CalledFus(Funktion->GetMaxEintrag()+1);
        // Berechnete Funktionen
        vector<bool> BerFus(Funktion->GetMaxEintrag()+1);
        bool modifieddiag = false;
        while (i < (Funktion->GetMaxEintrag()) && Erg)
        { // Alle Funktionen
          i++;

          Funktion->SetAkEint(i);
          if ((*Funktion)->Namen.Length() > 0)
          { // wenn Bezeichnung vorhanden
            if (!CalledFus[i])
            { // Funktion noch nicht berechnet
              // Holt Nummer der LETZTEN (UNTERSTEN) Funktion dieser Bezeichnung
              int tst = EFDoc->GetParser().BezeichnerFuNummer((*Funktion)->Namen.c_str());
              Funktion->SetAkEint(tst);
              MGPStr nm;
              AkFuNameOhneKlammern(nm);
              FuStr.SetStr(nm); // erzeuge Funktions-Bezeichner
              Ma->ParBezeichner(FuStr,Ma->GetFirst(),*YDaten);
              Ma->FunktionsBezeichner(FuStr,*YDaten);
              if ((_tcslen(FuStr.c_str()) > 0) &&
                  (tmp->FuStringVorh(FuStr.c_str())))
              { // wenn Graph mit dieser Bezeichnung vorhanden
                bool modified = UpdateAkDiag;
                ParserObj::FehlerRec Fehler;
                TEFFunktionCalcData FCD(Fehler);
                FCD.AkFeld = 0;
                FCD.TestModified = true; // flag fuer Modified-Flag ermitteln
                FCD.BerFus = &BerFus;
                FCD.DiagData = tmp->GetDiagData();
                TEFFunktionCalc FunktionCalc(&EFDoc->GetParser(),FCD);
                FunktionCalc.NextStep(); // 1. Wert berechnen fuer modified-Flag
                modified |= FCD.TestModified;
                if (modified)
                { // modifizierte Funktionen neu berechnen
                  if (!modifieddiag)
                    tmp->AddUndo(TEFGraphControlData::UndoDDiagData);
                  Erg = Berechnen(true,true,tmp);
                  modifiedany |= modified;
                  modifieddiag |= modified;
                }
                CalledFus[tst] = true;
                for (int t = i;t < tst;t++)
                { // Alle nachfolgenden Funktionen dieses Namens nicht berechnen
                  Funktion->SetAkEint(t);
                  MGPStr tmp1;
                  AkFuNameOhneKlammern(tmp1);
                  if (!_tcscmp(tmp1.c_str(),nm.c_str()))
                    CalledFus[t] = true;
                }
              }
            }
          }
        }
        for (size_t i = 0;i < BerFus.size();i++)
          GesBerFus[i] = GesBerFus[i] || BerFus[i];
        if (modifieddiag)
          tmp->UpdateWerBer();
      }
      if (UpdateAkDiag)
        child = 0;
      else
        child = child->GetNext();
    }
    if (modifiedany)
    { // Modifikationen in Editor eintragen
      Editor->SetLineModified(false,GesBerFus);
      EFDoc->SetIsDirty(true);
    }
    Funktion->SetAkEint(AkFunk);
    AkDiagW = OrgDiag;
    if (UpdateAkDiag)
      LastCalcDiagW = AkDiagW;
  }
  if (BerDlg)
    BerDlg->Destroy();
  BerDlg = 0;
}

void TEFView::OnUpdateGraph(wxCommandEvent &)
{
  if (AkDiagW)
  {
    GetFunctionsIncludingModules(Funktion,true);
    if (Editor->FindFocus() == AkDiagW)
    { // wenn Diagramm aktiv, die zug. Funktion des aktuellen Graphen berechnen
      const _TCHAR *tst = AkDiagW->GetAkFuString();
      if (tst)
      { // Graph vorhanden
        ParserObj *Psr = &EFDoc->GetParser();
        // Holt Nummer der LETZTEN (UNTERSTEN) Funktion dieser Bezeichnung
        int ind = Psr->BezeichnerFuNummer(tst);
        if (ind)
          Editor->SetSelection(Editor->GetLineIndex(ind-1),
                               Editor->GetLineIndex(ind-1));
        else
        { // Funktion nicht gefunden
          wxMessageBox(_("Graph specified as function parameter not found in current diagram !"),
                       wxT("Easy Funktion"),wxOK+wxICON_EXCLAMATION+wxCENTRE,
                       wxGetApp().GetTopWindow());
          return;
        }
      }
    }
    Berechnen(true,false);
  }
  if (BerDlg)
    BerDlg->Destroy();
  BerDlg = 0;
}

void TEFView::OnLabelEntry(wxCommandEvent &)
{
  if (!LabelEntryDlg)
  {
    TEFLabelEntryDlg *tmp = new TEFLabelEntryDlg(this);
    tmp->Create(wxGetApp().GetTopWindow(),dlgLabelEntry,_("Enter labels"));
    LabelEntryDlg = tmp;
  }
  LabelEntryDlg->Show(true);
}

void TEFView::OnEinzWert(wxCommandEvent &)
{
  wxWindow *wnd = GetFrame()->FindWindow(XRCID("pnlEinzelwert"));
  TEFEinzelwertPanel *pnl = (TEFEinzelwertPanel *)wnd;
  if (!pnl)
  {
    EinzelwertData ED;

    ED.Editor = Editor;
    ED.CD = EFDoc->GetEinzHistory();
    ED.EFView = this;
    ED.LDV = wxGetApp().GetLDV();
    pnl = new TEFEinzelwertPanel(ED);
    wxXmlResource::Get()->LoadPanel(pnl,GetFrame(),wxT("pnlEinzelwert"));
    pnl->Initialize();
  }
  else
  {
    pnl->Show(!pnl->IsShown());
    pnl = 0;
  }
  EFMDIChildFrame *frame = (EFMDIChildFrame *)GetFrame();
  frame->LayoutChildren(pnl);
  EFDoc->SetIsDirty(true);
}

void TEFView::OnSubmodules(wxCommandEvent &)
{
  wxWindow *wnd = GetFrame()->FindWindow(XRCID("pnlSubmodules"));
  TEFSubmodulesPanel *pnl = (TEFSubmodulesPanel *)wnd;
  if (!pnl)
  {
    SubmodulesData SD;

    SD.EFView = this;
    SD.m_submodules = ((TEFDocument *)GetDocument())->m_submodules;
    pnl = new TEFSubmodulesPanel(SD);
    wxXmlResource::Get()->LoadPanel(pnl,GetFrame(),wxT("pnlSubmodules"));
    pnl->Initialize();
  }
  else
  {
    pnl->Show(!pnl->IsShown());
    pnl = 0;
  }
  EFMDIChildFrame *frame = (EFMDIChildFrame *)GetFrame();
  frame->LayoutChildren(pnl);
  EFDoc->SetIsDirty(true);
}

void TEFView::OnBerechnen(wxCommandEvent &)
{
  if (Editor->FindFocus() != Editor)
  {
    wxMessageBox(_("Can only calculate graphs in function window !"),_("Error"),
                 wxOK+wxICON_EXCLAMATION,wxGetApp().GetTopWindow());
    return;
  }
  GetFunctionsIncludingModules(Funktion,true);
  if (Funktion->GetAkEint().Namen.Length() == 0)
  {
    wxTextEntryDialog dlg(Editor,_("Enter name of function"),
                                 _("Function without name"));
    dlg.ShowModal();
    wxString str = dlg.GetValue();
    if (str.length() > 0)
    {
      unsigned int ind = Editor->GetLineIndex(Funktion->AkEintNummer()-1);
      Editor->SetSelection(ind,ind);
      if (str[0u] != '[')
      {
        str.Prepend(wxT("["));
        str.Append(wxT("]"));
      }
      str.Append(wxT(" = "));
      Editor->Insert(str.c_str());
      GetFunctionsIncludingModules(Funktion,true);
    }
  }
  Berechnen(false,false);
  if (BerDlg)
    BerDlg->Destroy();
  BerDlg = 0;
}

bool TEFView::Berechnen(bool Update,bool AkFunk,TEFGraphControl *ThisDiag)
{
  int RetStatus;

  if (ThisDiag || AkDiagW)
  {
    ParserObj::FehlerRec tmpf;
    vector<bool> BerFus(Funktion->GetMaxEintrag()+1);
    BerechnenData Data;
    Data.EFView = this;
    Data.EFGraph = (ThisDiag ? ThisDiag : AkDiagW);
    AkFuNameOhneKlammern(Data.FuName);
    Data.AkFeld = 0;
    Data.Update = Update;
    Data.BerFus = (!AkFunk ? &BerFus : 0);
    if ((*Funktion)->ParamStr[0] != 0)
    { // wenn Parameter vorhanden, dazug. Funktion im akt. Diagramm suchen
      _TCHAR Zeile[200];
      int len = (*Funktion)->ParamStr.Length();
      // Fehlermeldung wenn Anfuehrungszeichen fehlen
      // -> Parameter statt Graph-Bezeichnung !
      if ((len < 5) || !((*Funktion)->ParamStr[1] == '\'') ||
          !((*Funktion)->ParamStr[len-2] == '\''))
      {
        tmpf.Art = ParserObj::SyntaxError;
        tmpf.Pos = 0;
        tmpf.FuNummer = Funktion->AkEintNummer();
        Fehlermeldung(tmpf,true);
        return false;
      }
      // Kopie von ParamStr ohne Klammern und '
      _tcscpy(Zeile,&(*Funktion)->ParamStr[2]);
      Zeile[_tcslen(Zeile)-2] = 0; // Hintere Klammer und ' loeschen
      // Daten des angegebenen Graphen suchen
      Data.AkFeld = (ThisDiag ? ThisDiag : AkDiagW)->FuStringVorh(Zeile);
      if (!Data.AkFeld)
      { // Fehlermeldung wenn Funktion nicht gefunden
        tmpf.Art = ParserObj::GraphNotFound;
        tmpf.Pos = 0;
        tmpf.FuNummer = Funktion->AkEintNummer();
        Fehlermeldung(tmpf,true);
        return false;
      }
    }
    if (!BerDlg)
    {
      wxString label;
      label.Printf(wxT("%s '%s'"),_("Calculate"),Data.FuName.c_str());
      TEFBerechnenDlg *tmp = new TEFBerechnenDlg(wxGetApp().GetTopWindow(),
                                                 dlgBerechnen,label,
                                                 EFDoc->GetParser());
      BerDlg = tmp;
    }
    if (!AkFunk) 
      (ThisDiag ? ThisDiag : AkDiagW)->AddUndo(TEFGraphControlData::UndoDDiagData);
    ((TEFBerechnenDlg *)BerDlg)->Init(Data);
    RetStatus = BerDlg->ShowModal();
    BerDlg->Show(false);
    BerDlg->Destroy();
    BerDlg = 0;
    if (Data.Fehler.Art != ParserObj::NoError)
    {
      Fehlermeldung(Data.Fehler,true);
    }
    if (!AkFunk)
    {
      (ThisDiag ? ThisDiag : AkDiagW)->UpdateWerBer();
      EFDoc->SetIsDirty(true);
      Editor->SetLineModified(false,BerFus);
    }
    return !((RetStatus == wxID_CANCEL) ||
             (Data.Fehler.Art != ParserObj::NoError));
  }
  else return false;
}

void TEFView::Fehlermeldung(ParserObj::FehlerRec &Fehler,bool FehlerPos,
                            TEFEdit *edit)
{
  int index,pos;
  wxString Zeile;

  if (!edit)
    edit = Editor;

  switch (Fehler.Art)
  {
   case ParserObj::SyntaxError:
        Zeile = _("Syntax error occured !");
        break;
   case ParserObj::GraphNotFound:
        Zeile = _("Graph specified as function parameter not found in current diagram !");
        break;
   case ParserObj::DivByZero:
        Zeile = _("Division by zero !");
        break;
   case ParserObj::WrongValue:
        Zeile = _("Value invalid !");
        break;
   case ParserObj::UnknownCommand:
        Zeile = _("Unknown function !");
        break;
   case ParserObj::InvalidCommand:
        Zeile = _("Function not allowed here !");
        break;
   case ParserObj::WrongSymbol:
        Zeile = _("Invalid character !");
        break;
   case ParserObj::Overflow:
        Zeile = _("Overflow !");
        break;
   case ParserObj::WrongLabel:
        Zeile = _("Unknown label !");
        break;
   case ParserObj::BracketError:
        Zeile = _("Parenthesis error !");
        break;
   case ParserObj::Domain:
        Zeile = _("Parameter out of valid range !");
        break;
   case ParserObj::Singularity:
        Zeile = _("Singularity error !");
        break;
   case ParserObj::Recursion:
        Zeile = _("Too many recursions !");
        break;
   case ParserObj::ParameterCount:
        Zeile = _("Invalid number of arguments !");
        break;
   case ParserObj::Convergence:
        Zeile = _("No convergence for integration !");
        break;
   case ParserObj::ImagResult:
        Zeile = _("Result is not defined for real numbers !");
        break;
   default :
        Zeile = _("Unknown error !");
  }
  if (Fehler.HasXValue)
  {
    MGPStr Z;
    EFDoc->GetParser().Ausgabe(Z,Fehler.XVal,wxGetApp().ZahlForm);
    wxString tmp;
    tmp.Printf(_("\nLast X value was %s."),Z.c_str());
    Zeile += tmp;
  }
  if ((Fehler.FuNummer > 0) && FehlerPos)
  {
    Zeile += _("\n\nShow error position ?");
    int ret = wxMessageBox(Zeile,wxT("Easy Funktion"),
                           wxYES_NO+wxICON_EXCLAMATION+wxCENTRE,
                           wxGetApp().GetTopWindow());
    if (ret != wxYES)
      return;
    Funktion->SetAkEint(Fehler.FuNummer);
    index = edit->GetLineIndex(Fehler.FuNummer-1);
    pos = Fehler.Pos;
    if ((*Funktion)->Namen[0] != 0)
    {
      pos += (*Funktion)->Namen.Length(); // Um NAMEN verschieben
      pos += (*Funktion)->ParamStr.Length(); // um Parameter verschieben
      pos += 3; // ' = ' bercksichtigen
    }
    edit->SetSelection(index+pos,index+pos); // Caret auf Fehler-Pos
    edit->SetFocus(); // Editor selektieren
  }
  else
    wxMessageBox(Zeile,wxT("Easy Funktion"),wxOK+wxICON_EXCLAMATION+wxCENTRE,
                 wxGetApp().GetTopWindow());
}

void TEFView::AddKlammern(FunctionArrObj *Fu,int Anf,int End)
{ // eckige Klammern um NAMEN-Eintraege eintragen
  MGPStr tmp;
  if (!End)
    End = Fu->GetMaxEintrag();
  for (int f = Anf;f <= End;f++)
  {
    Fu->SetAkEint(f);
    if ((Fu->GetAkEint().Namen.Length() > 0) &&
        (Fu->GetAkEint().Namen[0] != '[') &&
        (Fu->GetAkEint().Namen[0] > 32))
    {
      tmp.SetStr(wxT("["));
      tmp.AddStr(Fu->GetAkEint().Namen);
      tmp.AddStr(wxT("]"));
      Fu->GetAkEint().Namen.SetStr(tmp);
    }
  }
}

void TEFView::OnUmformen(wxCommandEvent &)
{
  int State;
  int i;
  MGPStr EStr;
  FunctionArrObj *Fu;
  unsigned int Anf,End;
  UmformenData UD;
  NumListFormat LDV;

  TEFEdit *edit = wxDynamicCast(Editor->FindFocus(),TEFEdit);
  if (!edit)
    edit = Editor;
  if (EFDoc->IsModified())
  { // Funktionen veraendert und nicht gespeichert
    int Res = wxOK;
    if (wxGetApp().Options & MyApp::keyWarnSolve)
      Res = wxMessageBox(_("Before solving functions the document should be saved. Save file now ?"),
                         wxT("Easy Funktion"),
                         wxYES_NO+wxICON_EXCLAMATION+wxCENTRE,
                         wxGetApp().GetTopWindow());
    if (Res == wxCANCEL)
      return;
    else
    if (Res == wxYES)
    { // Datei speichern
      wxCommandEvent event(wxID_SAVE,0);
      EFDoc->GetDocumentManager()->OnFileSave(event);
    }
  }
  EFDoc->AuflVorlage.Begin();
  ZahlFormatObj ZF(EFDoc->GetParser().GetLDV());
  do
  {
    ZF.StdToWin(EFDoc->AuflVorlage->BeibehaltenStr);
    ZF.StdToWin(EFDoc->AuflVorlage->ZusammenStr);
    ZF.StdToWin(EFDoc->AuflVorlage->ErgStr);
    EFDoc->AuflVorlage->ParamsList.Begin();
    if (EFDoc->AuflVorlage->ParamsList.GetMaxEintrag() >= 0)
      do
        ZF.StdToWin(EFDoc->AuflVorlage->ParamsList.GetAkEint());
      while (EFDoc->AuflVorlage->ParamsList.NextEint());
  }
  while (EFDoc->AuflVorlage.NextEint());
  UD.Vorlage = &EFDoc->AuflVorlage;
  UD.ErgH = EFDoc->GetErgHistory();
  UD.ParH = EFDoc->GetBeiHistory();
  UD.ZusH = EFDoc->GetZusHistory();
  edit->GetSelection(Anf,End);
  if (Anf != End) // wenn Text selektiert, setze auf selektierter Text
    UD.AkZeileOpt = 1;
  else
    UD.AkZeileOpt = 0;
  TEFUmformenDlg *U = new TEFUmformenDlg(UD);
  State = wxID_CANCEL;
  if (U)
  {
    U->Create(wxGetApp().GetTopWindow(),dlgUmformen,_("Solve functions"));
    State = U->ShowModal();
    U->Destroy();
  }
  EFDoc->AuflVorlage.Begin();
  EFDoc->AuflVorlage.GetAkEint() = UD.AkSett; // Einstellungen kopieren
  do // Strings ins Standart-Format umwandeln
  {
    ZF.WinToStd(EFDoc->AuflVorlage->BeibehaltenStr);
    ZF.WinToStd(EFDoc->AuflVorlage->ZusammenStr);
    ZF.WinToStd(EFDoc->AuflVorlage->ErgStr);
    EFDoc->AuflVorlage->ParamsList.Begin();
    if (EFDoc->AuflVorlage->ParamsList.GetMaxEintrag() >= 0)
      do
        ZF.WinToStd(EFDoc->AuflVorlage->ParamsList.GetAkEint());
      while (EFDoc->AuflVorlage->ParamsList.NextEint());
  }
  while (EFDoc->AuflVorlage.NextEint());
  if (State == wxID_OK)
  {
    // sicherstellen dass Leerzeile am Ende
    edit->GetSelection(Anf,End);
    if (edit->GetTextLen() > edit->GetLineIndex(edit->GetNumLines()-1))
    {
      edit->SetSelection(edit->GetTextLen(),edit->GetTextLen(),false);
      edit->Insert(wxT("\r\n"),false);
    }
    edit->SetSelection(Anf,End);
    EFDoc->AuflVorlage.Begin();
    // diese Referenz ist notwendig wg. Umwandlung mit WinToStd !
    AuflVorlageStruct &Eint0 = EFDoc->AuflVorlage.GetAkEint();
    bool found = false;
    while (EFDoc->AuflVorlage.NextEint())
    {
      if ((_tcscmp(UD.AkSett.Name.c_str(),
                   EFDoc->AuflVorlage->Name.c_str()) == 0))
      { // wenn Template bereits vorhanden, ueberschreiben
        found = true;
        EFDoc->AuflVorlage.GetAkEint() = Eint0;
      }
    }
    if (!found && (Eint0.Name[0])) // wenn neues Template mit Namen
    {
      EFDoc->AuflVorlage.NeuAkEint();
      EFDoc->AuflVorlage.GetAkEint() = Eint0;
    }
    Fu = new FunctionArrObj;
    if (Fu && GetFunktionen(Fu,false,edit))
    {
      if (edit == Editor)
        AddKlammern(Fu);
      EFDoc->AuflVorlage.Begin();
      EFDoc->AuflVorlage->ParamsList.Begin();
      int AnzParams = 0;
      if (EFDoc->AuflVorlage->ParamsList.GetMaxEintrag() >= 0)
        do
        { // wenn Definitionen vorhanden, am Anfang einfuegen (wg. Performance)
          Fu->SetAkEint(1);
          Fu->InsAkEint();
          MGPStr tmp(EFDoc->AuflVorlage->ParamsList.GetAkEint());
          ZF.StdToWin(tmp);
          ParseFunktion(tmp,Fu->GetAkEint(),edit);
          AnzParams++;
        }
        while (EFDoc->AuflVorlage->ParamsList.NextEint());
      if (AnzParams)
        AddKlammern(Fu,1,AnzParams);
      LDV = EFDoc->GetParser().GetLDV();
      unsigned short Mode;
      Mode = (unsigned short)(UD.AkSett.Mode*SolverObj::SolverTeilen +
                             (UD.AkSett.Anh & EFDaten::UDebug ?
                                              SolverObj::SolverDebug : 0) +
                             (UD.AkSett.Anh & EFDaten::UKlKuerzen ?
                                              SolverObj::SolverKlKuerzen : 0) +
                             (UD.AkSett.Anh & EFDaten::URules ?
                                              SolverObj::SolverRegelnOut : 0));
      TEFSolverDlg *SD = new TEFSolverDlg(Fu,UD.AkSett.ErgStr.c_str(),
                                          UD.AkSett.BeibehaltenStr.c_str(),
                                          UD.AkSett.ZusammenStr.c_str(),
                                          Mode,EFDoc->GetParser());
      if (SD)
      {
        SD->Create(wxGetApp().GetTopWindow(),dlgEdit,_("Solve functions"));
        State = SD->ShowModal();
        SD->Destroy();
        if (State == wxID_OK)
        { // Ergebnis einfuegen
          if (edit != Editor)
          {
            if (wxMessageBox(_("Insert results in function editor\ninstead of current equations editor ?\nGraphs can only be computed in function editor !"),
                             _("Easy Funktion"),
                             wxYES_NO+wxICON_QUESTION,
                             wxGetApp().GetTopWindow()) == wxYES)
              edit = Editor;
          }
          GetFunktionen(Funktion,true,edit);
          if (UD.AkZeileOpt) // Selektierter Text
          {
            edit->GetSelection(Anf,End);
            Anf = edit->GetLineFromPos(Anf);
            unsigned int End1 = End;
            End = edit->GetLineFromPos(End1);
            if (edit->GetLineIndex(End) < End1)
              End++;
          }
          else // Aktuelle Gleichung
          {
            Anf = edit->GetLineFromPos(-1);
            End = Anf+1;
          }
          if (UD.AkSett.Anh & EFDaten::UAnh) // Ergebnis anh�gen
            edit->SetSelection(edit->GetTextLen(),edit->GetTextLen());
          else // Gleichungen ersetzen
          {
            Anf = edit->GetLineIndex(Anf);
            End = edit->GetLineIndex(End);
            edit->SetSelection(Anf,End);
            edit->Cut(); // bisherige Funktionen ins ClipBoard
          }
          // Anfangsposition der neuen Funktionen
          Anf = edit->GetLineIndex(-1);
          if (UD.AkSett.Anh & EFDaten::UComment)
          {
            unsigned int Anf,End,tm;
            edit->GetSelection(Anf,tm);
            MGPStr tmp;
            tmp.AddStr(wxT(" "));
            tmp.AddStr(UD.AkSett.Name);
            tmp.AddStr(wxT(" "));
            if (EFDoc->AuflVorlage->ParamsList.GetMaxEintrag() >= 0)
            {
              tmp.AddStr(wxT("("));
              EFDoc->AuflVorlage->ParamsList.Begin();
              do
              {
                MGPStr t(EFDoc->AuflVorlage->ParamsList.GetAkEint());
                ZF.StdToWin(t);
                tmp.AddStr(t);
                if (EFDoc->AuflVorlage->ParamsList.AkEintNummer() <
                    EFDoc->AuflVorlage->ParamsList.GetMaxEintrag())
                  tmp.AddStr(wxGetApp().GetLDV().ListSeparator.c_str());
              }
              while (EFDoc->AuflVorlage->ParamsList.NextEint());
              tmp.AddStr(wxT(")"));
            }
            tmp.AddStr(wxT("\r\n"));
            if (tmp.Length() > 4)
            {
              edit->Insert(tmp.c_str());
              edit->GetSelection(tm,End);
              edit->SetSelection(Anf,End);
              edit->SetSelectionFormat(TEFTextCalc::Bold |
                                       TEFTextCalc::FontHeight,12);
              edit->SetSelection(End,End);
            }
          }
          MGPStr tmp;
          for (i = 1;i <= Fu->GetMaxEintrag();i++)
          {
            Funktion->NeuAkEint();
            Funktion->SetAkEint(Funktion->GetMaxEintrag());
            Fu->SetAkEint(i);
            // wenn anh�gen, nur BERECHNETE Gleichungen
            if (Fu->Berechnet() || !(UD.AkSett.Anh & EFDaten::UAnh))
            {
              tmp.AddStr(Fu->GetAkEint().Namen);
              tmp.AddStr(wxT(" = "));
              tmp.AddStr(Fu->GetAkEint().Term);
              tmp.AddStr(wxT("\r\n"));
            }
          }
          if (tmp.Length() > 0)
            edit->Insert(tmp.c_str());
          int line = edit->GetLineIndex(edit->GetLineFromPos(-1));
          edit->SetSelection(Anf,line);
        }
      }
      delete Fu;
    }
    EFDoc->SetIsDirty(true);
  }
  edit->SetFocus();
  EFDoc->AuflVorlage.Begin(); // 0 entry always without template name
  EFDoc->AuflVorlage->Name.SetStr(wxT(""));
}

void TEFView::OnNaeherung(wxCommandEvent &)
{
  MGPStr Zeile;
  int State;
  unsigned int Anf,End;
  TEFNaeherungDlg *U;
  MGPStr EStr;
  ZahlFormatObj ZF(EFDoc->GetParser().GetLDV());
  GraphManagerObj *YDaten;
  if (AkDiagW)
    YDaten = &AkDiagW->GetDiagData()->YDaten;
  else
    YDaten = 0;

  NaeherungDataType *NaehData = &EFDoc->NaehData;
  U = new TEFNaeherungDlg(*NaehData);
  State = wxID_CANCEL;
  if (U)
  {
    U->Create(wxGetApp().GetTopWindow(),
              dlgNaeherung,_("Approximation function"));
    State = U->ShowModal();
    U->Destroy();
  }
  if (State == wxID_OK)
  {
    EStr.SetStr(wxT(""));
    if (NaehData->Mode != 1)
    { // Polynom or approximation function
      FuGenObj FG(EFDoc->GetParser().GetLDV());
      if (NaehData->m_CurGraph && YDaten && YDaten->YDatenVorhanden())
      {
        size_t i = 0;
        size_t count = 0;
        if (YDaten->AkGraph().GetPunkte() < NaehData->m_NumPts)
          NaehData->m_NumPts = YDaten->AkGraph().GetPunkte();
        while (count < NaehData->m_NumPts)
        {
          // Exit if last point in graph is exceeded
          if (i > YDaten->AkGraph().GetPunkte()-1)
            break;

          // Make sure that last point is not less than last point in graph
          if (count == NaehData->m_NumPts-1)
            i = YDaten->AkGraph().GetPunkte()-1;

          FG.Ausgabe(Zeile,YDaten->AkGraph().GetX(i),
                      ZahlFormatObj::FitForm);
          EStr.AddStr(Zeile);
          EStr.AddStr(EFDoc->GetParser().GetLDV().ListSeparator.c_str());
          FG.Ausgabe(Zeile,real(YDaten->AkGraph().GetY(i)),
                      ZahlFormatObj::FitForm);
          EStr.AddStr(Zeile);
          EStr.AddChar(' ');
          size_t alti = i;
          count++;
          
          // calculate next point in graph
          i = size_t(floor(double(YDaten->AkGraph().GetPunkte()-1)*count/(NaehData->m_NumPts-1)+0.5));

          // Avoid duplicate points
          if (alti == i)
            i = alti + 1;
        }
        NaehData->ParamStr.SetStr(EStr);
      }
      else
        EStr.SetStr(NaehData->ParamStr);
      if (FG.ReadPkts(EStr.c_str()))
      {
        if (NaehData->Mode == FuGenObj::Polynom)
          EStr.SetStr(FG.Interpol());
        else
          EStr.SetStr(FG.Ausgleich(NaehData->Mode));
      }
      else
      {
        wxMessageBox(_("Couldn't read points !"),wxT("Easy Funktion"),
                     wxOK+wxICON_EXCLAMATION+wxCENTRE,
                     wxGetApp().GetTopWindow());
        return;
      }
    }
    else
    { // Exponential function
      double y1 = ZF.ZahlWert(NaehData->YStr[0].c_str());
      double y2 = ZF.ZahlWert(NaehData->YStr[1].c_str());
      double y3 = ZF.ZahlWert(NaehData->YStr[2].c_str());
      double x1 = ZF.ZahlWert(NaehData->XStr[0].c_str());
      double x2 = ZF.ZahlWert(NaehData->XStr[1].c_str());
      if (((fabs(y3-y1) <= 2*fabs(y2-y1)*1.001) &&
          (fabs(y3-y1) >= 2*fabs(y2-y1)*0.999)) || (fabs(x2-x1) < 1e-30))
      { // invalid entries
        wxMessageBox(_("Function couldn't be generated !"),wxT("Easy Funktion"),
                     wxOK+wxICON_EXCLAMATION+wxCENTRE,
                     wxGetApp().GetTopWindow());
        return;
      }
      else
      {
        double B = (y2-y1)*(y2-y1)/(y3-y1-2*(y2-y1));
        double tmp = 0;
        if (fabs(B) > 1e-30) tmp = (y2-y1+B)/B;
        if ((fabs(B) > 1e-30) && (tmp > 1e-30))
        {
          double A = y1 - B;
          double C = log(tmp)/(x2-x1);
          ZF.Ausgabe(Zeile,A,
                     wxGetApp().ZahlForm == ZahlFormatObj::IngForm ?
                     ZahlFormatObj::FitForm : ZahlFormatObj::ExpForm);
          EStr.SetStr(Zeile);
          ZF.Ausgabe(Zeile,B,
                     wxGetApp().ZahlForm == ZahlFormatObj::IngForm ?
                     ZahlFormatObj::FitForm : ZahlFormatObj::ExpForm);
          if (B >= 0)
            EStr.AddChar('+');
          EStr.AddStr(Zeile);
          ZF.Ausgabe(Zeile,C,wxGetApp().ZahlForm == ZahlFormatObj::IngForm ?
                     ZahlFormatObj::FitForm : ZahlFormatObj::ExpForm);
          EStr.AddStr(wxT("*exp("));
          EStr.AddStr(Zeile);
          EStr.AddStr(wxT("*(x"));
          ZF.Ausgabe(Zeile,-x1,wxGetApp().ZahlForm == ZahlFormatObj::IngForm ?
                     ZahlFormatObj::FitForm : ZahlFormatObj::ExpForm);
          if (-x1 >= 0)
            EStr.AddChar('+');
          EStr.AddStr(Zeile);
          EStr.AddStr(wxT("))"));
        }
      }
    }
    if (EStr[0] != 0) // append result
    {
      GetFunktionen(Funktion,true);
      Editor->SetSelection(Editor->GetTextLen(),
                           Editor->GetTextLen());
      Editor->GetSelection(Anf,End);
      MGPStr tmp;
      tmp.AddStr(wxT("\r\n"));
      if (NaehData->ErgStr[0] != 0)
      {
        tmp.AddChar('[');
        tmp.AddStr(NaehData->ErgStr.c_str());
        tmp.AddStr(wxT("] = "));
      }
      while (EStr[0] == ' ')
        EStr.DelNChar(0,1);
      tmp.AddStr(EStr.c_str());
      tmp.AddStr(wxT("\r\n"));
      Editor->Insert(tmp.c_str());
      Editor->SetSelection(Anf,
                           Editor->GetLineIndex(Editor->GetLineFromPos(-1)));
    }
    else
    {
      wxMessageBox(_("Function couldn't be generated !"),wxT("Easy Funktion"),
                   wxOK+wxICON_EXCLAMATION+wxCENTRE,wxGetApp().GetTopWindow());
      return;
    }
    EFDoc->SetIsDirty(true);
  }
}

void TEFView::OnFunktionen(wxCommandEvent &)
{
  FunktionsData Data;

  wxWindow *wnd = GetFrame()->FindWindow(XRCID("pnlFunktionen"));
  TEFFunktionsPanel *pnl = (TEFFunktionsPanel *)wnd;
  if (!pnl)
  {
    Data.Mode = wxGetApp().Options;
    Data.EFView = this;
    pnl = new TEFFunktionsPanel(Data);
    wxXmlResource::Get()->LoadPanel(pnl,GetFrame(),wxT("pnlFunktionen"));
    pnl->Initialize();
  }
  else
  {
    pnl->Show(!pnl->IsShown());
    pnl = 0;
  }
  EFMDIChildFrame *frame = (EFMDIChildFrame *)GetFrame();
  frame->LayoutChildren(pnl);
  EFDoc->SetIsDirty(true);
}

void TEFView::OnInteractiveSolver(wxCommandEvent &)
{
#ifdef EFMULTITHREAD
  InteractiveSolverData Data;

  wxWindow *wnd = GetFrame()->FindWindow(XRCID("pnlInteractiveSolver"));
  TEFInteractiveSolverPanel *pnl = (TEFInteractiveSolverPanel *)wnd;
  if (!pnl)
  {
    Data.Mode = 1;
    Data.View = this;
    Data.Parser = &EFDoc->GetParser();
    pnl = new TEFInteractiveSolverPanel(Data);
    wxXmlResource::Get()->LoadPanel(pnl,GetFrame(),wxT("pnlInteractiveSolver"));
    TEFLabelComboBox *lcb = new TEFLabelComboBox(pnl,XRCID("ID_EXPRESSION"),this);
    wxXmlResource::Get()->AttachUnknownControl(wxT("ID_EXPRESSION"),lcb);
    pnl->Fit();
  }
  else
  {
    pnl->Show(!pnl->IsShown());
    pnl = 0;
  }
  EFMDIChildFrame *frame = (EFMDIChildFrame *)GetFrame();
  frame->LayoutChildren(pnl);
  EFDoc->SetIsDirty(true);
#endif
}

void TEFView::InsertString(const _TCHAR *zeile,int Anf,int End)
{ // fuegt zeile ein, markiert Anf bis End
  wxWindow *wnd = GetFrame()->FindWindow(XRCID("pnlEinzelwert"));
  TEFEinzelwertPanel *pnl = (TEFEinzelwertPanel *)wnd;
  if (pnl && pnl->IsShown() && pnl->GetLocalEditorActive())
  {
    pnl->InsertString(zeile,Anf,End);
  }
  else
  {
    if (!AkEdit)
      AkEdit = Editor;
    unsigned int AAnf,AEnd;
    AkEdit->GetSelection(AAnf,AEnd);
    AkEdit->Insert(zeile);
    AkEdit->SetSelection(Anf+AAnf,End+AAnf);
    AkEdit->SetFocus();
  }
}

void TEFView::CeSave(wxUpdateUIEvent &event)
{
  event.Enable(EFDoc && EFDoc->IsModified());
}

void TEFView::OnSendTo(wxCommandEvent &)
{
  if (!EFDoc || !EFDoc->Save())
    return;
#if defined(__BORLANDC__) || defined(__MINGW32__) || defined(__VISUALC__)
  wxString name,ext;
  wxFileName::SplitPath(EFDoc->GetFilename(),0,0,&name,&ext);
  name += wxT(".")+ext;
  wxMailMessage msg(name,wxGetUserName(),wxT(""),wxT(""),
                    EFDoc->GetFilename(),name);
  wxEmail em;
  if (!em.Send(msg))
#endif
    wxMessageBox(_("Couldn't send eMail!"));
}

void TEFView::OnHomepage(wxCommandEvent &)
{
  wxString tmp = wxT("http://easyfunktion.sourceforge.net/");
  if (!wxLaunchDefaultBrowser(tmp))
    wxMessageBox(_("Could not open Website!"),tmp);
}

void TEFView::OnForum(wxCommandEvent &)
{
  if (!wxLaunchDefaultBrowser(wxT("http://de.groups.yahoo.com/group/EasyFunktion")))
    wxMessageBox(_("Could not open Website!"));
}

void TEFView::OnFeedback(wxCommandEvent &)
{
  wxString title = wxT("mailto:markusgreither@yahoo.de?subject=");
  title += _("User feedback Easy Funktion ");
  title += wxT(VERSIONSTRING);
  if (!wxLaunchDefaultBrowser(title))
    wxMessageBox(wxT("Couldn't send eMail."));
}

void TEFView::OnPlayMacro(wxCommandEvent &)
{
  wxString name = wxFileSelector(_("Select macro name"),wxT(""),
                                 wxT("macro.xml"),wxT("xml"),
                                 wxT("*.xml"),
                                 wxFD_OPEN+wxFD_FILE_MUST_EXIST,
                                 Editor);
  if (!name.IsEmpty())
  {
    if (MyMacroRecorder::Get()->PlayMacro(name.mb_str()))
    {
      wxCommandEvent evt1(wxEVT_COMMAND_MENU_SELECTED,
                          XRCID("ICM_MACRO"));
      AddPendingEvent(evt1);
    }
    else
      wxMessageBox(_("Error reading macro!"),name,
                   wxOK+wxICON_EXCLAMATION+wxCENTRE,
                   wxGetApp().GetTopWindow());
  }
}

void TEFView::OnMacro(wxCommandEvent &)
{
  bool nextmacro = true;
  EFXmlNode *node = MyMacroRecorder::Get()->GetAkNode();
  if (node)
  {
    const MGPStr &name = node->GetName();
    wxCommandEvent evt(wxEVT_COMMAND_MENU_SELECTED,
                       EFXmlResource::Get()->GetXRCID(name.c_str()));
    evt.SetClientData(node);
    if (MyMacroRecorder::Get()->IsStepping())
      wxMessageBox(_("Press OK for next step"));
    while (wxGetApp().Pending())
      wxGetApp().Dispatch();
    if (node->GetProperties())
    {
      EFXmlProperty *prop = node->GetProperties();
      int ret;
      if (MyMacroRecorder::Get()->GetPropertyInt(prop,wxT("window"),ret))
      {
        wxWindow *wnd = Editor->FindWindow(ret);
        if (wnd)
        {
          wnd->SetFocus();
	  evt.SetEventObject(wnd);
	}
      }
      if (MyMacroRecorder::Get()->GetPropertyInt(prop,wxT("checked"),ret))
        evt.SetInt(ret);
    }
    if (!wxStrcmp(node->GetName().c_str(),wxT("ICM_SUBMACRO")))
    {
      EFXmlNode *text = node->GetChildren();
      wxString tmp;
      if (text)
        tmp = text->GetContent().c_str();
      if (!MyMacroRecorder::Get()->PlaySubMacro(tmp.mb_str()))
        wxMessageBox(_("Error reading macro!"),tmp,
                     wxOK+wxICON_EXCLAMATION+wxCENTRE,
                     wxGetApp().GetTopWindow());
    }
    else
    {
      ProcessEvent(evt);
      MyMacroRecorder::Get()->GetNextNode();
    }
  }
  else
    nextmacro = !MyMacroRecorder::Get()->EndMacro();
  if (nextmacro)
  {
    while (wxGetApp().Pending())
      wxGetApp().Dispatch();
    wxCommandEvent evt1(wxEVT_COMMAND_MENU_SELECTED,XRCID("ICM_MACRO"));
    AddPendingEvent(evt1);
  }
}

void TEFView::OnRecordMacro(wxCommandEvent &)
{
  MyMacroRecorder::Get()->RecordMacro();
}

void TEFView::OnEndMacro(wxCommandEvent &)
{
  wxString name;
  if (MyMacroRecorder::Get()->IsRecording())
    name = wxFileSelector(_("Select macro name"),wxT(""),
                          wxT("macro.xml"),wxT("xml"),
                          wxT("*.xml"),
                          wxFD_SAVE+wxFD_OVERWRITE_PROMPT,
                          Editor);
  MyMacroRecorder::Get()->EndMacro(name.mb_str());
}

void TEFView::CeMacroActive(wxUpdateUIEvent &event)
{
  event.Enable(MyMacroRecorder::Get()->IsPlaying() ||
               MyMacroRecorder::Get()->IsRecording());
}
void TEFView::CeMacroInactive(wxUpdateUIEvent &event)
{
  event.Enable(!MyMacroRecorder::Get()->IsPlaying() &&
               !MyMacroRecorder::Get()->IsRecording());
}

void TEFView::OnMacroEditFont(wxCommandEvent &event)
{
  TEFEdit *edit = wxDynamicCast(wxWindow::FindFocus(),TEFEdit);
  if (!edit)
    return;

  EFXmlNode *node = (EFXmlNode *)event.GetClientData();
  FontType font;
  EFDoc->XmlLoadEditFont(node,font);
  SetEditorFont(edit,font);
  if (edit == Editor)
    EFDoc->Editor = font;
  EFDoc->SetIsDirty(true);
}
void TEFView::OnRenumberChildren(wxCommandEvent &)
{
  RenumberChildWindows();
}

void TEFView::OnClearModified(wxCommandEvent &)
{
  Editor->ClearModify();
  EFDoc->SetIsDirty(false);
}

void TEFView::RenumberChildWindows()
{
  if (!Editor)
    return;
  size_t EditCount = 0,DiagCount = 0,PictureCount = 0;
  Editor->ClearModify();
  wxWindowListNode *Child = Editor->GetChildren().GetFirst();
  while (Child)
  {
    TEFEquationsControl *edit = wxDynamicCast((wxWindow *)Child->GetData(),
                                              TEFEquationsControl);
    TEFGraphControl *diag = wxDynamicCast((wxWindow *)Child->GetData(),
                                          TEFGraphControl);
    wxResizeableControl *res = wxDynamicCast((wxWindow *)Child->GetData(),
                                             wxResizeableControl);
    if (edit)
    {
      EditCount++;
      edit->GetEditor()->SetId(ID_EDITOR+EditCount);
      edit->SetId(ID_EQUATIONS+EditCount);
    }
    else
    if (diag)
    {
      DiagCount++;
      diag->SetId(ID_DIAGRAMM+DiagCount);
    }
    else
    if (res)
    {
      PictureCount++;
      ((wxWindow *)Child->GetData())->SetId(ID_PICTURE+PictureCount);
    }
    Child = Child->GetNext();
  }
}

void TEFView::CeInteractiveSolver(wxUpdateUIEvent &event)
{
  wxWindow *frame = GetFrame();
  if (frame)
    event.Check(frame->FindWindow(XRCID("pnlInteractiveSolver")));
}

void TEFView::CeFunktionen(wxUpdateUIEvent &event)
{
  wxWindow *frame = GetFrame();
  if (frame)
    event.Check(frame->FindWindow(XRCID("pnlFunktionen")));
}

void TEFView::CeEinzWert(wxUpdateUIEvent &event)
{
  wxWindow *frame = GetFrame();
  if (frame)
    event.Check(frame->FindWindow(XRCID("pnlEinzelwert")));
}

void TEFView::CeSubmodules(wxUpdateUIEvent &event)
{
  wxWindow *frame = GetFrame();
  if (frame)
    event.Check(frame->FindWindow(XRCID("pnlSubmodules")));
}

BEGIN_MACRO_EVENT_TABLE(TEFView, wxView, MyMacroRecorder)
  // Macro events
  EVT_MENU(XRCID("ICM_MACRO"), TEFView::OnMacro)
  EVT_MENU(MACROID("CMD_DELWINDOW"), TEFView::OnDelWindow)
  EVT_MENU(MACROID("CMD_EDITFONT"), TEFView::OnMacroEditFont)

  // Clipboard events
  EVT_MENU(WXMACROID(wxID_PASTE), TEFView::OnEditPaste)
  EVT_UPDATE_UI(wxID_PASTE, TEFView::CeEditPaste)

  EVT_MENU(WXMACROID(wxID_COPY), TEFView::OnFocusWindowEvent)
  EVT_UPDATE_UI(wxID_COPY, TEFView::CeFocusWindowEvent)

  EVT_MENU(WXMACROID(wxID_CUT), TEFView::OnFocusWindowEvent)
  EVT_UPDATE_UI(wxID_CUT, TEFView::CeFocusWindowEvent)

  EVT_MENU(MACROID("CMD_COPYBRACKETS"), TEFView::OnFocusWindowEvent)
  EVT_UPDATE_UI(MACROID("CMD_COPYBRACKETS"), TEFView::CeFocusWindowEvent)

  // General child window notification events
  EVT_MENU(XRCID("ICM_DELWINDOW"), TEFView::OnDelWindow) // for window delete
  EVT_MENU(XRCID("ICM_SETMODIFIED"), TEFView::OnSetModified)

  // Editor notification events
  EVT_MENU(XRCID("ICM_RENUMBERCHILDREN"), TEFView::OnRenumberChildren)
  EVT_MENU(XRCID("ICM_EDITSETFOCUS"), TEFView::OnEditSetFocus)

  // Diagram notification events
  EVT_MENU(XRCID("ICM_GETFUNKTIONS"), TEFView::OnGetFunktions)
  EVT_MENU(XRCID("ICM_SETDIAGMODIFIED"), TEFView::OnSetDiagModified)
  EVT_MENU(XRCID("ICM_SETAKDIAG"), TEFView::OnSetAkDiag)

  // Print events
  EVT_MENU(wxID_PREVIEW, TEFView::OnPreview)
  EVT_MENU(wxID_PRINT, TEFView::OnPrintFocusWindow)
  EVT_MENU(wxID_PRINT_SETUP, TEFView::OnPageSetup)

  // General commands
  EVT_UPDATE_UI(XRCID("CMD_DELDIAG"), TEFView::CeDelDiag)

  EVT_MENU(MACROID("CMD_PAGEBREAK"), TEFView::OnPageBreak)
  EVT_UPDATE_UI(XRCID("CMD_PAGEBREAK"), TEFView::CeAnyEdit)

  EVT_MENU(MACROID("CMD_UPDATEALLGRAPHS"), TEFView::OnUpdateAll)

  EVT_MENU(MACROID("CMD_NEWDIAGRAM"), TEFView::OnNeuAkDiag)
  EVT_UPDATE_UI(XRCID("CMD_NEWDIAGRAM"), TEFView::CeEditor)

  EVT_MENU(MACROID("CMD_NEWEQUATIONS"), TEFView::OnNeuEquations)
  EVT_UPDATE_UI(XRCID("CMD_NEWEQUATIONS"), TEFView::CeEditor)

  EVT_MENU(MACROID("CMD_NEWGRAPH"), TEFView::OnBerechnen)
  EVT_UPDATE_UI(XRCID("CMD_NEWGRAPH"), TEFView::CeEditor)

  EVT_MENU(MACROID("CMD_UPDATEGRAPH"), TEFView::OnUpdateGraph)
  EVT_UPDATE_UI(XRCID("CMD_UPDATEGRAPH"), TEFView::CeEditor)

  // Not macro recorded general events
  EVT_MENU(XRCID("ICM_CLEARMODIFIED"), TEFView::OnClearModified)

  EVT_MENU(XRCID("ICM_LABELENTRY"), TEFView::OnLabelEntry)
  EVT_UPDATE_UI(XRCID("ICM_LABELENTRY"), TEFView::CeAnyEdit)

  EVT_MENU(XRCID("ICM_EDITFONT"), TEFView::OnEditFont)
  EVT_UPDATE_UI(XRCID("ICM_EDITFONT"), TEFView::CeAnyEdit)

  EVT_MENU(XRCID("ICM_ZOOM"), TEFView::OnZoom)

  EVT_MENU(XRCID("ICM_SELECTFUNCTION"), TEFView::OnFunktionen)
  EVT_UPDATE_UI(XRCID("ICM_SELECTFUNCTION"), TEFView::CeFunktionen)

  EVT_MENU(XRCID("ICM_INTERACTIVESOLVER"), TEFView::OnInteractiveSolver)
  EVT_UPDATE_UI(XRCID("ICM_INTERACTIVESOLVER"), TEFView::CeInteractiveSolver)

  EVT_MENU(XRCID("ICM_SUBMODULES"), TEFView::OnSubmodules)
  EVT_UPDATE_UI(XRCID("ICM_SUBMODULES"), TEFView::CeSubmodules)

  EVT_MENU(XRCID("ICM_SOLVER"), TEFView::OnUmformen)
  EVT_UPDATE_UI(XRCID("ICM_SOLVER"), TEFView::CeAnyEdit)

  EVT_MENU(XRCID("ICM_APPROXIMATIONFUNCTION"), TEFView::OnNaeherung)
  EVT_MENU(XRCID("ICM_SINGLEVALUE"), TEFView::OnEinzWert)
  EVT_UPDATE_UI(XRCID("ICM_SINGLEVALUE"), TEFView::CeEinzWert)

  EVT_UPDATE_UI(wxID_SAVE,TEFView::CeSave)

  EVT_MENU(wxID_DUPLICATE, TEFView::OnSendTo)
  EVT_MENU(XRCID("ICM_FEEDBACK"), TEFView::OnFeedback)
  EVT_MENU(XRCID("ICM_HOMEPAGE"), TEFView::OnHomepage)
  EVT_MENU(XRCID("ICM_FORUM"), TEFView::OnForum)
  EVT_MENU(XRCID("ICM_WERTBERECHNEN"), TEFView::OnWertBerechnen)

  EVT_MENU(XRCID("ICM_RECORDMACRO"), TEFView::OnRecordMacro)
  EVT_UPDATE_UI(XRCID("ICM_RECORDMACRO"), TEFView::CeMacroInactive)

  EVT_MENU(XRCID("ICM_ENDMACRO"), TEFView::OnEndMacro)
  EVT_UPDATE_UI(XRCID("ICM_ENDMACRO"), TEFView::CeMacroActive)

  EVT_MENU(XRCID("ICM_PLAYMACRO"), TEFView::OnPlayMacro)
  EVT_UPDATE_UI(XRCID("ICM_PLAYMACRO"), TEFView::CeMacroInactive)

  // Undo/Redo events
  EVT_MENU(WXMACROID(wxID_UNDO), TEFView::OnFocusWindowEvent)
  EVT_UPDATE_UI(wxID_UNDO, TEFView::CeFocusWindowEvent)

  EVT_MENU(WXMACROID(wxID_REDO), TEFView::OnFocusWindowEvent)
  EVT_UPDATE_UI(wxID_REDO, TEFView::CeFocusWindowEvent)
END_EVENT_TABLE()

